import requests
import traceback
import time
import numpy as np
from sqlalchemy.orm import sessionmaker
from ...api.raw import *
from .raw_data_helper import RawDataHelper
from ...view.raw_models import OSFundNav
from ...api.raw import RawDataApi

class OverseaDataUpdate:

    def __init__(self, update_period:int=35, data_per_page:int=20, data_helper=RawDataHelper()):
        self.update_period = update_period
        self.data_per_page = data_per_page
        self.data_helper = data_helper
        self.raw_data_api = RawDataApi()
        self.today = datetime.datetime.now().date()

    def update_unit(self, fund_id, exsited_price, fund_id_list):
        nav_dic = {
                'navDate':'datetime',
                'price':'nav',
            }
        url = 'https://fund.bluestonehk.com/fund/ifast/info/getHistoryNavs?productId={}&pageSize={}&pageNo={}'.format(
            fund_id, self.data_per_page, 1
        )
        response = requests.get(url)
        datas = response.json()
        if datas['message'] != '请求成功' or datas['body']['data'] == []:
            return None
        new_nav = pd.DataFrame(datas['body']['data'])[nav_dic.keys()].rename(columns=nav_dic)
        td = pd.to_datetime(new_nav.datetime)
        td = [_.date() for _ in td]
        new_nav.datetime = td
        exsited_date = exsited_price[exsited_price['codes'] == fund_id].datetime.max()
        new_df = new_nav[new_nav.datetime > exsited_date].copy()
        if new_df.empty:
            idx = fund_id_list.index(fund_id)
            if idx % 10 == 0:
                time.sleep(0.5)
                print(f'finish fund nav {idx}')
            return None
        new_df.loc[:,'codes'] = fund_id
        idx = fund_id_list.index(fund_id)
        if idx % 10 == 0:
            time.sleep(0.5)
            print(f'finish fund nav {idx}')
        return new_df

    def update_fund_nav(self):
        failed_tasks = []
        try:
            fund_info = self.raw_data_api.get_over_sea_fund_info()
            fund_id_list = fund_info.codes.tolist()
            update_period_dates = self.today - datetime.timedelta(days=self.update_period)
            exsited_price = self.raw_data_api.get_oversea_fund_nav(start_date = update_period_dates).drop(columns=['_update_time'], errors='ignore')
            result = []
            for fund_id in fund_id_list:
                _count = 0
                while True:
                    try:
                        if _count == 5:
                            break                        
                        new_df = self.update_unit(fund_id, exsited_price, fund_id_list)
                        #print(f'    fund_id {fund_id}')
                        if new_df is not None:
                            result.append(new_df)
                    except Exception as e:
                        _count += 1
                        print(e)
                        time.sleep(1)
                    else:
                        break
            if len(result) != 0:
                nav_result = pd.concat(result)
                self.data_helper._upload_raw(nav_result, OSFundNav.__table__.name)
                

        except Exception as e:
            print(e)
            traceback.print_exc()
            failed_tasks.append(f'oversea_fund_nav')
        return failed_tasks
    
    def update_price_unit(self, fund_id, fund_id_list):
        nav_dic = {
                'navDate':'datetime',
                'price':'nav',
            }
        url = 'https://fund.bluestonehk.com/fund/ifast/info/getHistoryNavs?productId={}&pageSize={}&pageNo={}'.format(
            fund_id, self.data_per_page, 1
        )
        response = requests.get(url)
        datas = response.json()
        if datas['message'] != '请求成功' or datas['body']['data'] == []:
            return None
        new_df = pd.DataFrame(datas['body']['data'])
        if new_df.empty:
            idx = fund_id_list.index(fund_id)
            if idx % 10 == 0:
                time.sleep(0.5)
                print(f'finish fund nav {idx}')
            return None
        idx = fund_id_list.index(fund_id)
        if idx % 10 == 0:
            time.sleep(0.5)
            print(f'finish fund nav {idx}')
        return new_df

    def update_fund_adj_factor(self):
        failed_tasks = []
        try:
            fund_info = self.raw_data_api.get_over_sea_fund_info()
            fund_id_list = fund_info.codes.tolist()
            nav_div_factor = self.raw_data_api.get_os_fund_nav_div_factor().drop(columns=['_update_time'], errors='ignore')
            result = []
            for fund_id in fund_id_list:
                _count = 0
                while True:
                    try:
                        if _count == 5:
                            break   
                        idx = fund_id_list.index(fund_id)
                        new_nav_ori = self.update_price_unit(fund_id, fund_id_list)
                        if new_nav_ori is None:
                                print(f'    fund_id {fund_id} does not have nav idx {idx}')
                                break
                        if (new_nav_ori.adjPrice / new_nav_ori.price).unique().tolist() == [1]:
                            print(f'    fund_id {fund_id} does not have divident idx {idx}')
                            break
                        new_nav = new_nav_ori.rename(columns={'navDate':'datetime','adjPrice':'before_adj_price'}).set_index('datetime').sort_index()[['before_adj_price','price']]
                        td = pd.to_datetime(new_nav.index)
                        td = [i.date() for i in td]
                        new_nav.index = td
                        adj_factor_df = nav_div_factor[nav_div_factor.product_id==fund_id][['ex_date','factor']].set_index('ex_date')
                        adj_factor_df.index.name = 'datetime'
                        new_nav.loc[:,'factor'] = new_nav.before_adj_price / new_nav.price
                        new_nav = new_nav[['factor']]
                        new_nav = new_nav.round(4).drop_duplicates().iloc[1:]
                        new_nav = new_nav / new_nav.iloc[0]
                        _adj_factor_df = adj_factor_df.loc[new_nav.index[0]:]
                        new_nav = new_nav * _adj_factor_df.iloc[0]
                        con = new_nav.index > _adj_factor_df.index[-1]
                        if not np.any(con):
                            print(f'    fund_id {fund_id} does not have new adj factor idx {idx}')
                            break
                        new_nav = new_nav[con].copy()
                        new_nav = new_nav.reset_index().rename(columns={'index':'ex_date'})
                        new_nav.loc[:,'product_id'] = fund_id
                        print(f'    fund_id {fund_id} idx {idx} total {len(fund_id_list)} finish')
                        if new_nav is not None:
                            result.append(new_nav)
                    except Exception as e:
                        _count += 1
                        print(e)
                        time.sleep(1)
                    else:
                        break
            df_adj_new = pd.concat(result)
            self.data_helper._upload_raw(df_adj_new, OSFundNAVDivFactor.__table__.name)
            return True
        except Exception as e:
            print(e)
            traceback.print_exc()
            failed_tasks.append(f'oversea_fund_nav')
        return False

    def update_index_price(self):
        # todo change file location here
        folder = '/Users/huangkejia/Downloads/Data/'
        excel_name = '8.0_海外基金基准_210220_回溯2000版本_unlink.xlsx'
        sheet_name = 'Benchmark dailydata'
        df = pd.read_excel(folder+excel_name,sheet_name=sheet_name).replace(dic)
        with RawDatabaseConnector().managed_session() as db_session:        
            query = db_session.query(
                OSIndexPrice
            )
            index_price_exsited = pd.read_sql(query.statement, query.session.bind)
        index_price_exsited = index_price_exsited.drop(columns=['_update_time'], errors='ignore')
        dic = {'VNIndex':'VNINDEX','dax':'DAX'}
        fund_cols = df.columns.tolist()
        fund_col_len = len(fund_cols)
        result = []
        for i in range(fund_col_len):
            if i % 2 == 0:
                _fund_cols = fund_cols[i:i+2]
                _df = df[_fund_cols]
                _index_id = _df[_fund_cols[0]].values[0].replace(' Index', '').replace(' index', '')
                _df = _df.iloc[3:].dropna()
                if _df.empty:
                    continue
                _df.columns = ['datetime','close']
                con = [ not isinstance(_, int) for _ in _df.datetime]
                _df = _df[con]
                td = pd.to_datetime(_df.datetime)
                td = [i.date() for i in td]
                _df.datetime = td
                if _index_id in dic.keys():
                    _index_id = dic[_index_id]
                _df.loc[:,'codes'] = _index_id
                _df_exsited = index_price_exsited[index_price_exsited.codes == _index_id]
                _df_new = _df[~_df.datetime.isin(_df_exsited.datetime)]
                result.append(_df_new)
        df_result = pd.concat(result).drop_duplicates(subset=['datetime','codes'])
        df_result.to_sql('oversea_index_price', RawDatabaseConnector().get_engine(), index=False, if_exists='append')

    def check_adj_factor(self):
        # 复权因子有问题的代码
        fund_adj_fac = RawDataApi().get_os_fund_nav_div_factor().drop(columns='_update_time')
        product_id_list = fund_adj_fac.product_id.unique()
        problem_fund_id = []
        for product_id in product_id_list:
            _df = fund_adj_fac[fund_adj_fac.product_id == product_id].dropna()
            if _df.empty:
                continue
            _df.loc[:,'next_single_factor'] = _df['single_factor'].shift(-1)
            _df.loc[:,'calc_factor'] = _df.factor * _df.next_single_factor
            _df.loc[:,'next_factor'] = _df['factor'].shift(-1)
            _df.loc[:,'rate'] = abs(_df['calc_factor'] / _df['next_factor'] - 1)
            if round(_df.rate.max(),4) > 0:
                #print(_df)
                problem_fund_id.append(product_id)

        # 正确的数据
        change_result = []
        for product_id in problem_fund_id:
            _df = fund_adj_fac[fund_adj_fac.product_id == product_id].dropna()
            _df.loc[:,'next_single_factor'] = _df['single_factor'].shift(-1)
            _df.loc[:,'calc_factor'] = _df.factor * _df.next_single_factor
            _df.loc[:,'next_factor'] = _df['factor'].shift(-1)
            _df_not_null = _df.dropna().copy()
            if _df.empty:
                continue
            fac = _df.factor.values[0]
            res = [fac]
            for next_i in _df_not_null.next_single_factor:
                fac = fac * next_i
                res.append(fac)
            _df.loc[:,'right_factor'] = res
            _df.loc[:,'rate'] = abs(_df['right_factor'] / _df['factor'] - 1)
            _df = _df[_df.rate.round(4) > 0].copy()
            change_result.append(_df)

        # 替换正确的数据
        right_df = pd.concat(change_result)[['product_id','ex_date','factor']]
        change_fund_list = right_df.product_id.unique().tolist()
        session = sessionmaker(RawDatabaseConnector().get_engine())
        db_session = session()
        for row in db_session.query(OSFundNAVDivFactor).filter(OSFundNAVDivFactor.product_id.in_(change_fund_list)).all():
            target_df = right_df[(right_df.product_id == row.product_id) & (right_df.ex_date == row.ex_date)]
            if target_df.empty:
                continue
            target = target_df.factor.values[0]
            #print(f'{row.product_id} {row.ex_date} {target}')
            row.factor = float(target)
        db_session.commit()
        db_session.close()  


import datetime
import pandas as pd
import numpy as np
import traceback
from typing import List, Dict
from ...api.raw import RawDataApi
from ..raw.raw_data_helper import RawDataHelper
from ....data.view.raw_models import QSOverseaFundCurMdd, QSOverseaFundMonthlyRet, QSOverseaFundIndicator, QSOverseaFundRadarScore, QSOverseaFundPeriodRet, OSFundRet, QSFundRecentRet
from ....util.calculator_item import CalculatorBase

# TODO 按照一只美股持仓基金和中国股票持仓基金跑流程
TEST_CODE = ['HKAACE01','HKAAVP'] # 净值分析 持仓分析 选用的基金id
TEST_CODE_COMPARE = ['MXAPJ', 'MXWD', 'MXCNANM', 'RIY'] # 因子计算截面对比 选用的指数id


class OverseaFundNavAnalysis:

    # 对每个基金类别统计,按照数量和含义人工汇总
    FUND_GROUP = {
        'stock':['股票基金','指数基金 - 股票'],
        'fix_rate':['定息基金','指数基金 - 定息'],
        'mix':['均衡基金','混合资产基金'],
        'other':['另类基金'],
        'mmf':['货币市场基金'],
      }
    
    # 统计每类基金类别下最普遍基准作为默认基准
    DEFAULT_BENCHMARK = {
        '股票基金':'MXAPJ',
        '定息基金':'JPEIGLBL',
        '均衡基金':'MXAPJ',
        '另类基金':'mmf',
        '混合资产基金':'SHSZ300',
        '指数基金 - 股票':'MXAPJ',
        '指数基金 - 定息':'JPEIGLBL',
        '货币市场基金':'mmf',
    }

    # 资产基准
    ASSET_BENCHMARK = {
        'us_stock': 'RIY',      # 罗素1000指数
        'us_bond': 'LUACTRUU',  # Bloomberg Barclays US Corporate Total Return Value Unhedged USD
        'ap_stock': 'MXAP', # MSCI AC Asia ex Japan Net Total Return USD Index
        'ap_bond': 'JPEIJACC',  # J.P. Morgan Asia Credit Index Core
    }

    NOT_CONSIDER_TYPE = ['指数基金']

    def __init__(self, data_helper: RawDataHelper):
        self._data_helper = data_helper
        self._raw_api = RawDataApi()

    def init(self, end_date=str):
        fund_info = self._raw_api.get_over_sea_fund_info()
        fund_benchmark = self._raw_api.get_over_sea_fund_benchmarks()
        self.fund_info = pd.merge(fund_info, fund_benchmark[['benchmark_final','industry_1','industry_2','isin_code']], on='isin_code').set_index('codes')    
        self.fund_info = self.fund_info[~self.fund_info.fund_type.isin(self.NOT_CONSIDER_TYPE)]
        for fund_type, benchmark in self.DEFAULT_BENCHMARK.items():
            fund_list = self.fund_info[(self.fund_info['fund_type'] == fund_type) & (self.fund_info.benchmark_final.isnull())].index.tolist()
            self.fund_info.loc[fund_list,'benchmark_final'] = benchmark
        self._fund_nav = self._raw_api.get_oversea_fund_nav(end_date=end_date).drop(columns=['_update_time'],errors='ignore')
        self.fund_nav = self._fund_nav.pivot_table(index='datetime',columns='codes',values='nav').replace(0,np.nan).ffill()
        self.fund_nav_adj = self._raw_api.get_oversea_fund_nav_adj(end_date=end_date)
        self.fund_nav_adj = self.fund_nav_adj.pivot_table(index='datetime',columns='codes',values='nav').replace(0,np.nan).ffill()
        self.index_price = self._raw_api.get_oversea_index_price(end_date=end_date)        
        self.index_price = self.index_price.pivot_table(index='datetime',columns='codes',values='close').replace(0,np.nan).ffill()
        self.get_new_fund_benchmark()

    def get_new_fund_benchmark(self):
        # 由于数据更新 当前把基金分四类设置基准
        us_stock_fund_list = self.fund_info[(self.fund_info.area.isin(['环球','美国'])) 
                      &(self.fund_info.industry_1 == 'Equity Fund')].index.tolist()
        us_bond_fund_list = self.fund_info[(self.fund_info.area == '美国')&(self.fund_info.industry_1 == 'Debt Fund')].index.tolist()
        ap_stock_fund_list = self.fund_info[((self.fund_info.area == '亚洲 (包括日本)')|(self.fund_info.area == '亚洲 (日本除外)'))&(self.fund_info.industry_1 == 'Equity Fund')].index.tolist()
        ap_bond_fund_list = self.fund_info[((self.fund_info.area == '亚洲 (包括日本)')|(self.fund_info.area == '亚洲 (日本除外)')) & (self.fund_info.industry_1 == 'Debt Fund')].index.tolist()


        l = us_stock_fund_list+us_bond_fund_list+ap_stock_fund_list+ap_bond_fund_list
        fl = self.fund_info.index.tolist()
        fls = [i for i in fl if i not in l]

        _fund_info = self.fund_info.loc[fls]
        cl = _fund_info[_fund_info.fund_type == '股票基金'].area.unique().tolist()
        asian_country = ['泰国', '日本', '大中华', '新兴市场', '中国','新兴市场','中国A股','香港','印尼','台湾', '印度','印尼','中国及印度','越南', '菲律宾','前沿市場','马来西亚及新加坡','韩国',
        '新加坡']
        all_country = [i for i in cl if i not in asian_country]
        ap_stock_fund_list.extend(_fund_info[(_fund_info.fund_type == '股票基金') & (_fund_info.area.isin(asian_country))].index.tolist())
        us_stock_fund_list.extend(_fund_info[(_fund_info.fund_type == '股票基金') & (_fund_info.area.isin(all_country))].index.tolist())

        cl = _fund_info[_fund_info.fund_type != '股票基金'].area.unique().tolist()
        asian_country = ['中国A股','大中华','亚洲 (日本除外)','香港','新兴市场','中国','印度']
        all_country = [i for i in cl if i not in asian_country]
        us_bond_fund_list.extend(_fund_info[(_fund_info.fund_type != '股票基金') & (_fund_info.area.isin(all_country))].index.tolist())
        ap_bond_fund_list.extend(_fund_info[(_fund_info.fund_type != '股票基金') & (_fund_info.area.isin(asian_country))].index.tolist())

        benchmark_dic = {
            'us_stock':us_stock_fund_list,
            'us_bond':us_bond_fund_list,
            'ap_stock':ap_stock_fund_list,
            'ap_bond':ap_bond_fund_list,
        }
        fund_benchmark = {}
        for asset_id, fund_list in benchmark_dic.items():
            for fund_id in fund_list:
                fund_benchmark[fund_id] = asset_id
        self.fund_benchmark_dic = fund_benchmark


    def get_update_date(self, df_new, df_old, dt_col='datetime'): # dt_col in other db maybe report_date
        codes_list = df_new.codes.unique().tolist()
        res = []
        for code_i in codes_list:
            exsited_dt = df_old[df_old.codes == code_i]
            if exsited_dt.empty:
                continue
            exsited_dt = pd.to_datetime(exsited_dt.datetime.values[-1]).date()
            _df = df_new[(df_new.codes == code_i)&(df_new.datetime > exsited_dt)]
            # idx = codes_list.index(code_i)
            # if idx % 50 == 0:
            #     print(idx)
            res.append(_df)
        return pd.concat(res)

    def rolling_cur_mdd(self, x):
        if pd.isnull(x).all():
            return np.nan
        x_max = np.fmax.accumulate(x, axis=0)
        return -(1 - np.nanmin(x[-1] / x_max))

    def period_ret_calc(self, df, rule='1M'):
        df = df.set_axis(pd.to_datetime(df.index), inplace=False).resample(rule).last()
        df.index = [i.date() for i in df.index]
        df.index.name = 'date'
        fund_id = df.columns.tolist()[0]
        df.columns=['ret']
        df = df.pct_change(1).reset_index()
        if rule == '1M':
            df.date = [i.year * 100 + i.month for i in df.date]
        else:
            df.date = [i.year for i in df.date]
        df.loc[:,'codes'] = fund_id
        return df.dropna(axis=0)

    def process_fund_ret(self):
        failed_tasks: List[str] = []
        try:
            _df = self.fund_nav.pct_change(1).iloc[1:] * 100
            _df = pd.DataFrame(_df.stack())
            _df.columns=['ret']
            _df = _df.reset_index()
            self._data_helper._upload_raw(_df, OSFundRet.__table__.name,to_truncate=True)
        except Exception as e:
            print(e)
            traceback.print_exc()
            failed_tasks.append(f'oversea_fund_ret')
        return failed_tasks
        
    def process_current_mdd(self):
        failed_tasks: List[str] = []
        try:
            _df = (self.fund_nav / self.fund_nav.cummax()) - 1
            _df = pd.DataFrame(_df.stack())
            _df.columns=['current_mdd']
            _df = _df.reset_index()
            _df = _df[_df.datetime >= datetime.date(2021,1,1)]
            end_date = self.fund_nav.index[-1]
            existed_start_date = end_date - datetime.timedelta(days=30)
            exsited_data = self._raw_api.get_oversea_fund_cur_mdd(start_date = str(existed_start_date))
            _df = self.get_update_date(df_new=_df, df_old=exsited_data, dt_col='datetime')
            self._data_helper._upload_raw(_df, QSOverseaFundCurMdd.__table__.name)
        except Exception as e:
            print(e)
            traceback.print_exc()
            failed_tasks.append(f'oversea_current_mdd')
        return failed_tasks

    def process_monthly_ret(self):
        failed_tasks: List[str] = []
        try:
            _df = self.fund_nav.set_axis(pd.to_datetime(self.fund_nav.index), inplace=False).resample('1M').last()
            _df.index = [i.date() for i in _df.index]
            _df.index.name = 'datetime'
            _df = _df.pct_change(1)
            _df = pd.DataFrame(_df.stack())
            _df.columns=['monthly_ret']
            _df = _df.reset_index()
            _df = _df.replace(np.Inf,np.nan).replace(-np.Inf,np.nan)
            self._data_helper._upload_raw(_df, QSOverseaFundMonthlyRet.__table__.name, to_truncate=True)
        except Exception as e:
            print(e)
            traceback.print_exc()
            failed_tasks.append(f'oversea_monthly_ret')
        return failed_tasks

    def process_indicators(self):

        dic = {
            'cumu_ret':'累计收益',
            'annual_ret':'年化收益',
            'annual_vol':'年化波动',
            'sharpe':'夏普',
            'recent_1w_ret':'最近一个周收益',
            'recent_1m_ret':'最近一个月收益',
            'recent_3m_ret':'最近三个月收益',
            'recent_6m_ret':'最近六个月收益',
            'recent_1y_ret':'最近一年收益',
            'recent_3y_ret':'最近三年收益',
            'recent_5y_ret':'最近五年收益',
            'worst_3m_ret':'最差三个月收益',
            'worst_6m_ret':'最差六个月收益',
            'last_mv_diff':'最新涨跌幅',
            'last_increase_rate':'最新涨跌',
            'recent_drawdown':'当前回撤',
            'recent_mdd_date1':'当前回撤开始时间',
            'recent_mdd_lens':'当前回撤持续天数',
            'mdd':'最大回撤',
            'mdd_date1':'最大回撤开始日',
            'mdd_date2':'最大回撤结束日',
            'mdd_lens':'最大回撤持续天数',
            'calmar':'卡玛比率',
            'downside_risk':'下行风险',
            'sortino':'索提诺',
            'var':'风险价值',
            'ervar':'风险价值调整比',
            'skew':'偏度',
            'kurtosis':'峰度',
            'alpha_cl':'选股收益',
            'beta_cl':'择时收益',
            'beta_bear':'熊市beta', 
            'beta_bull':'牛市beta',
            'excess_ret':'相对收益',
            'win_rate':'胜率',
            'corr':'相关性',
            'track_err':'跟踪误差',
            'info':'信息比率',
            'alpha':'alpha',
            'beta':'beta',
            'treynor':'特雷诺',
            'not_system_risk':'非系统性风险',
            'cvar':'条件风险价值',
            'up_capture':'上行捕捉率',
            'down_capture':'下行捕捉率',
            'excess_ret_cotinue':'超额收益持续性',
            'continue_value_m':'业绩持续性',
        }

        failed_tasks: List[str] = []
        try:
            res = []
            monthly_indicator = []
            daily_indicator = ['start_date','end_date','trade_year','last_unit_nav','cumu_ret','annual_ret','annual_vol','sharpe','recent_1w_ret','recent_1m_ret','recent_3m_ret','recent_6m_ret','recent_1y_ret','recent_3y_ret','recent_5y_ret','worst_3m_ret','worst_6m_ret','last_mv_diff','last_increase_rate','recent_drawdown','recent_mdd_date1','recent_mdd_lens','mdd','mdd_date1','mdd_date2','mdd_lens']
            fund_list = self.fund_info.index.tolist()
            for fund_id in fund_list:
                if fund_id in ['HKATAF01','HKJFCNG'] or fund_id not in self.fund_nav:
                    continue
                fund_nav_i = self.fund_nav[[fund_id]]
                fund_nav_adj_i = self.fund_nav_adj[[fund_id]]
                benchmark_id = self.fund_info.loc[fund_id].benchmark_final
                #benchmark_id = self.ASSET_BENCHMARK[self.fund_benchmark_dic[fund_id]]
                
                if not benchmark_id in self.index_price or benchmark_id in ['USC0TR03','G0O1','I02885JP']:
                    benchmark_id = self.DEFAULT_BENCHMARK[self.fund_info.loc[fund_id,'fund_type']]
                fund_nav_i = fund_nav_i.join(self.index_price[benchmark_id]).ffill().dropna()
                fund_nav_i.columns = ['fund','benchmark']
                fund_nav_adj_i = fund_nav_adj_i.join(self.index_price[benchmark_id]).ffill().dropna()
                fund_nav_adj_i.columns = ['fund','benchmark']
                last_day = fund_nav_i.index.values[-1]
                # 不同年度
                for year in [1,3,5]:
                    b_d = last_day - datetime.timedelta(days=365*year)
                    _fund_nav_i = fund_nav_i.loc[b_d:]
                    _fund_nav_adj_i = fund_nav_adj_i.loc[b_d:]
                    if _fund_nav_i.fund.pct_change().std() < 0.0005:
                        continue
                    res_i_m = CalculatorBase.get_stat_result(
                            dates = _fund_nav_adj_i.index.values,
                            values = _fund_nav_adj_i.fund.values,
                            risk_free_rate=0.015,
                            frequency='1M',
                            ret_method='pct_ret',
                            benchmark_values=_fund_nav_adj_i.benchmark.values,
                        )
                    res_i_d = CalculatorBase.get_stat_result(
                                    dates = _fund_nav_i.index.values,
                                    values = _fund_nav_i.fund.values,
                                    risk_free_rate=0.015,
                                    frequency='1D',
                                    ret_method='pct_ret',
                                    benchmark_values=_fund_nav_i.benchmark.values,
                                )
                    _df_i_daily = pd.DataFrame([res_i_d])
                    _df_i_monthly = pd.DataFrame([res_i_m])
                    # 日度和月度结合
                    if monthly_indicator == []:
                        monthly_indicator = [i for i in _df_i_daily.columns if i not in daily_indicator]
                    _df_i = pd.concat([_df_i_daily[daily_indicator], _df_i_monthly[monthly_indicator]],axis=1)
                    _df_i.loc[:,'codes'] = fund_id
                    _df_i.loc[:,'data_cycle'] = year
                    res.append(_df_i)
                idx = fund_list.index(fund_id)
                print(f'fund {fund_id} {idx + 1}')
            _df = pd.concat(res)
            _df = _df.drop(columns=['start_date']).rename(columns={'end_date':'datetime'})
            _df = _df.replace(np.Inf,np.nan).replace(-np.Inf,np.nan)
            total_columns = list(QSOverseaFundIndicator.__table__.columns.keys())
            _df = _df[_df.columns.intersection(total_columns)]
            self._data_helper._upload_raw(_df, QSOverseaFundIndicator.__table__.name)
        except Exception as e:
            print(e)
            traceback.print_exc()
            failed_tasks.append(f'oversea_indicators')
        return failed_tasks

    def process_radar_score(self):
        failed_tasks: List[str] = []
        try:
            indicator_df = self._raw_api.get_qs_fund_indicator()
            dt = indicator_df.datetime.unique().tolist()[0]
            dic = {
                    'annual_ret':'ret_ability',
                    'annual_vol':'risk_ability',
                    'up_capture':'bull_ability',
                    'down_capture':'bear_ability',
                    'alpha':'alpha_ability',
                    'mdd':'drawdown_ability',}
            indicator_df = indicator_df[indicator_df.data_cycle == 3].drop(columns=['data_cycle']).set_index('codes')
            _df = indicator_df.copy()
            _df = _df[list(dic.keys())].rename(columns=dic)
            _df['risk_ability'] = - _df['risk_ability']
            _df['bear_ability'] = - _df['bear_ability']
            _df['drawdown_ability'] = - _df['drawdown_ability']
            _df = _df.rank(pct=True)
            _df['total_score'] = _df.mean(axis=1).rank(pct=True)
            _df = (_df / 0.2).apply(np.ceil)
            _df = _df.reset_index()
            _df.loc[:,'datetime'] = dt
            type_dic = {}
            for group_id, fund_types in self.FUND_GROUP.items():
                fund_list = self.fund_info[self.fund_info.fund_type.isin(fund_types)].index.tolist()
                for i in fund_list:
                    type_dic[i] = fund_types[0]
            _df.loc[:,'fund_type'] = _df.codes.map(lambda x:type_dic[x])
            type_dic = {
                'ret_ability':int,
                'risk_ability':int,
                'bull_ability':int,
                'bear_ability':int,
                'alpha_ability':int,
                'drawdown_ability':int,
                'total_score':int,
            }
            _df = _df.astype(type_dic)
            self._data_helper._upload_raw(_df, QSOverseaFundRadarScore.__table__.name)
        except Exception as e:
            print(e)
            traceback.print_exc()
            failed_tasks.append(f'oversea_radar_score')
        return failed_tasks
    
    def process_period_ret(self):
        failed_tasks: List[str] = []
        try:
            res = []
            for fund_id in self.fund_nav:
                fund_nav_i = self.fund_nav[[fund_id]]
                # monthly
                res_monthly = self.period_ret_calc(fund_nav_i, rule='1M')
                # yearly
                res_yearly = self.period_ret_calc(fund_nav_i, rule='1Y')
                res.append(res_monthly)
                res.append(res_yearly)
            df = pd.concat(res, axis=0)
            self._data_helper._upload_raw(df, QSOverseaFundPeriodRet.__table__.name, to_truncate=True)
        except Exception as e:
            print(e)
            traceback.print_exc()
            failed_tasks.append(f'oversea_period_ret')
        return failed_tasks

    def process_recent_ret(self):
        failed_tasks: List[str] = []
        try:
            res = []
            end_date = self.fund_nav.index[-1]
            for fund_id in self.fund_nav:
                nav = self._fund_nav[self._fund_nav.codes == fund_id].set_index('datetime')[['nav']].dropna()
                ret_i = CalculatorBase.calc_recent_ret(dates=nav.index,values=nav.nav.values)
                ret_i['datetime'] = end_date
                ret_i['codes'] = fund_id
                res.append(ret_i)
            df = pd.DataFrame(res)
            df = df.replace(np.Inf,np.nan).replace(-np.Inf, np.nan)
            df = df.replace(np.Inf,np.nan).replace(-np.Inf, np.nan)
            total_columns = list(QSFundRecentRet.__table__.columns.keys())
            df = df[df.columns.intersection(total_columns)]
            self._data_helper._upload_raw(df, QSFundRecentRet.__table__.name, to_truncate=True)
        except Exception as e:
            print(e)
            traceback.print_exc()
            failed_tasks.append(f'oversea_recent_ret')
        return failed_tasks

    def daily_process(self, end_date):
        self.init(end_date)
        failed_tasks = []
        failed_tasks.extend(self.process_fund_ret())
        #failed_tasks.extend(self.process_current_mdd())
        failed_tasks.extend(self.process_monthly_ret())
        failed_tasks.extend(self.process_period_ret())
        failed_tasks.extend(self.process_recent_ret())
        return failed_tasks

    def weekly_process(self, end_date):
        self.init(end_date)
        failed_tasks = []
        failed_tasks.extend(self.process_indicators())
        failed_tasks.extend(self.process_radar_score())
        return failed_tasks


from cassandra.cqlengine import columns
from cassandra.cqlengine.models import Model
from ...wrapper.cas.cas_connection import CassandraConnector
from ...wrapper.cas.constants import CAS_KEY_SPACE

CassandraConnector().get_conn()

class FundEstimatedNav(Model):
    __table_name__ = 'fund_estimated_nav'
    __keyspace__ = CAS_KEY_SPACE
    __connection__ = 'surfing'

    fund_id = columns.Text(primary_key=True)            # 基金ID
    timestamps = columns.List(columns.DateTime())       # 预估数据时间
    unit_net_values = columns.List(columns.Double)      # 预估净值
    acc_net_values = columns.List(columns.Double)       # 预估累计净值
    adjusted_net_values = columns.List(columns.Double)  # 预估后复权净值

    def to_dict(self):
        return {
            'fund_id': self.fund_id,
            'timestamps': self.timestamps,
            'unit_net_values': self.unit_net_values,
            'acc_net_values': self.acc_net_values,
            'adjusted_net_values': self.adjusted_net_values
        }






import traceback

from .basic_data_part1 import BasicDataPart1
from .basic_data_part2_fund_ret import BasicFundRet
from .style_analysis_data import StockFactor
from .basic_data_helper import BasicDataHelper
from .barra_cne5_factor_data import BarraCNE5FactorData


class BasicDataProcessor(object):
    def __init__(self):
        self._data_helper = BasicDataHelper()
        self.basic_data_part1 = BasicDataPart1(self._data_helper)
        self.basic_fund_ret = BasicFundRet(self._data_helper)
        self._style_analysis = StockFactor(self._data_helper)
        self._barra_cne5_factor = BarraCNE5FactorData(self._data_helper)

    def process_all(self, start_date, end_date):
        failed_tasks = []

        try:
            failed_tasks.extend(self.basic_data_part1.process_all(start_date, end_date))
        except Exception as e:
            print(e)
            traceback.print_exc()
            failed_tasks.append('unknown in basic_data_part1')

        try:
            failed_tasks.extend(self.basic_fund_ret.process_all(end_date))
            failed_tasks.extend(self._style_analysis.process_all(start_date, end_date))
            failed_tasks.extend(self._barra_cne5_factor.process_all(start_date, end_date))
        except Exception as e:
            print(e)
            traceback.print_exc()
            failed_tasks.append('unknown in other basic_data')

        return failed_tasks

    def get_updated_count(self):
        return self._data_helper._updated_count

    def process_not_trading_day(self, start_date, end_date):
        failed_tasks = []

        try:
            failed_tasks.extend(self.basic_data_part1.process_not_trading_day(start_date, end_date))
        except Exception as e:
            print(e)
            traceback.print_exc()
            failed_tasks.append('unknown in basic_data_part1')

        return failed_tasks


if __name__ == "__main__":
    BasicDataProcessor().process_all('20200813', '20200813')



import pandas as pd
import numpy as np
import json
import datetime
import re
from ...manager import DataManager
from ...wrapper.mysql import ViewDatabaseConnector
from ...view.view_models import FundRecommendationCollection
from .__init__ import mng_s3_uri, fund_s3_uri

EHR_PROBLEM_EKYWORDS = {
    'E': ['ETF', 'REITs', 'ESG', 'CES'],
    'H': ['H股', 'A-H50', '50AH', '中证AH'],
    'R': ['REITs'],
}

class FundRecommendationCollectionProcessor(object):

    def bench_parse(self, x):
        if x is None:
            return []
        bench_dic = json.loads(x)
        index_list = []
        for k,v in bench_dic.items():
            if k == '-1' or v < 0.4:
                continue
            if k == 'hs300' and v < 0.65:
                continue
            index_list.append(k)
        return index_list

    def append_data(self, table_name, data_append_directly_data_df):
        if not data_append_directly_data_df.empty:
            with ViewDatabaseConnector().managed_session() as mn_session:
                try:
                    mn_session.execute(f'TRUNCATE TABLE {table_name}')
                    mn_session.commit()
                except Exception as e:
                    print(f'Failed to truncate table {table_name} <err_msg> {e}')
            data_append_directly_data_df.to_sql(table_name, ViewDatabaseConnector().get_engine(), index = False, if_exists = 'append')
            print('新数据已插入')
        else:
            print('没有需要插入的新数据')
    
    def calc_whole_data(self):
        sector_info = DataManager.basic_data(func_name='get_sector_info')
        sector_fund = DataManager.basic_data(func_name='get_sector_index_info')
        fund_info = DataManager.basic_data(func_name='get_fund_info')
        index_info = DataManager.basic_data(func_name='get_index_info')
        benchmark_info = DataManager.basic_data(func_name='get_fund_benchmark')
        fund_ipo_stats = DataManager.basic_data(func_name='get_fund_ipo_stats')
        fund_open_info = DataManager.basic_data(func_name='get_fund_open_info')
        fund_status = DataManager.basic_data(func_name='get_fund_status')
        fund_conv_bond = DataManager.basic_data(func_name='get_fund_conv_stats')

        conv_bond_fund_list = fund_conv_bond[fund_conv_bond.conv_weights > 20].fund_id.tolist()
        abs_fund_ret_list = fund_info[fund_info.wind_class_2 == '股票多空'].fund_id.tolist()

        fund_info['trim_name'] = fund_info.desc_name.map(lambda x : re.subn(r'[ABCDEFR]{1,2}(\(人民币\)|\(美元现汇\)|\(美元现钞\)|1|2|3)?$', '', x)[0])
        self.not_include_in_market_fund_list = DataManager.basic_data(func_name='get_inside_market_funds').fund_id.tolist()
        not_include_close_fund_list = fund_open_info.fund_id.tolist()
        fund_info_money_fund = fund_info[fund_info.wind_class_2 == '货币市场型基金'].copy()
        self.not_include_no_buy_funds = fund_info_money_fund[ (fund_info_money_fund.desc_name.str.contains('B'))
                    | (fund_info_money_fund.desc_name.str.contains('C'))
                    | (fund_info_money_fund.desc_name.str.contains('D'))
                    | (fund_info_money_fund.desc_name.str.contains('E'))
                    | (fund_info_money_fund.desc_name.str.contains('F'))].fund_id.tolist()

        index_info = index_info.set_index('index_id')
        benchmark_info = benchmark_info[benchmark_info.index_text.notna()]
        benchmark_info.loc[:,'index_list'] = benchmark_info.benchmark_s_raw.map(lambda x : self.bench_parse(x))
        benchmark_info.loc[:,'index_name'] = benchmark_info.index_list.map(lambda x : index_info.loc[x].desc_name.tolist())
        sector_info = sector_info.set_index('sector_id')

        sector_fund = sector_fund.dropna(subset=['fund_id']).set_index('fund_id')
        fund_id_list = list(sector_fund.index.unique())
        res = []
        for fund_id in fund_id_list:
            _df = sector_fund.loc[[fund_id]]
            sector_list = list(_df.sector_id.dropna().unique())
            index_list = list(_df.index_id.unique())
            dic = {
                'fund_id': fund_id,
                'sector_list':sector_list,
                'sector_name':sector_info.loc[sector_info.index.intersection(sector_list)].sector_name.tolist()
            }
            res.append(dic)

        fund_ipo_stats = fund_ipo_stats.pivot_table(index='end_date', columns='fund_id', values='ipo_allocation_weight').fillna(0)
        fund_ipo_stats = fund_ipo_stats + fund_ipo_stats.shift(1).fillna(0) + fund_ipo_stats.shift(2).fillna(0) + fund_ipo_stats.shift(3).fillna(0)
        _df = fund_ipo_stats.tail(1).T
        _df.columns = ['ipo_weight']
        ipo_fund_list = _df[_df > 0.15].dropna().index.tolist()
            
        fund_info_part = fund_info[['fund_id','order_book_id','desc_name','company_id','wind_class_1','wind_class_2','risk_level']].set_index('fund_id')
        fund_info_part.loc[:,'is_ipo_fund'] = fund_info_part.index.isin(ipo_fund_list)
        sector_part = pd.DataFrame(res).set_index('fund_id')
        index_part = benchmark_info.set_index('fund_id')[['index_list','index_name']]

        fund_indicator_score = pd.read_parquet(fund_s3_uri)
        fund_info = fund_info[fund_info.fund_id.isin(fund_indicator_score.reset_index().dropna(subset=['total_score']).fund_id)]
        self.ehr_funds = []
        # 这里itertuples会比iterrows快1个数量级
        for r in fund_info[['desc_name', 'fund_id']].itertuples():
            for type_word, problem_words in EHR_PROBLEM_EKYWORDS.items():
                i = r.desc_name
                for problem_i in problem_words:
                    i = i.replace(problem_i, '')
                if type_word in i:
                    self.ehr_funds.append(r.fund_id)

        fund_info_no_money_fund = fund_info[fund_info.wind_class_2 != '货币市场型基金'].copy()
        fund_info_no_money_fund = fund_info_no_money_fund.groupby('trim_name').last()
        self.include_c_type = fund_info_no_money_fund.fund_id.tolist()
        
        
        fund_indicator_1 = fund_indicator_score[fund_indicator_score['fund_type'] != 'index']
        fund_indicator_2 = fund_indicator_score[fund_indicator_score.fund_type == 'index'].copy()
        #fund_indicator_2['ret_ability'] = fund_indicator_2.alpha_ability
        fund_indicator = pd.concat([fund_indicator_1,fund_indicator_2],axis=0)
        fund_indicator_part = fund_indicator[['mdd~history_daily','annual_ret~history_m','ret_ability','risk_ability','stable_ability','select_time','select_stock','mng_score','total_score','fund_type']]

        mng_info = DataManager.derived_data(func_name='get_fund_manager_info')
        mng_indicator_score = pd.read_parquet(mng_s3_uri)
        mng_part = mng_info[mng_info.end_date > datetime.datetime.now().date()].set_index('mng_id')[['fund_id']]
        mng_indicator_score = mng_indicator_score.join(mng_part, on='mng_id')
        mng_indicator_score_part = mng_indicator_score.rename(columns={'total_score':'mng_total_score'})[['mng_total_score','fund_id','mng_id','name','fund_type']].dropna(subset=['mng_total_score'])

        fund_type_list = list(fund_indicator_score.fund_type.unique())
        result = []
        fund_indicator_part = fund_indicator_part.drop(columns='mng_score')
        for fund_type in fund_type_list:
            _fund_indicator_part = fund_indicator_part[fund_indicator_part.fund_type == fund_type][['total_score']].reset_index().set_index('fund_id')
            _mng_indicator_part = mng_indicator_score_part[mng_indicator_score_part.fund_type == fund_type].set_index('fund_id')
            _fund_indicator_part = _fund_indicator_part.join(_mng_indicator_part)
            idx = _fund_indicator_part.groupby(['fund_id'], sort=False)['mng_total_score'].transform(max) == _fund_indicator_part.mng_total_score
            result.append(_fund_indicator_part[idx][['mng_id','name','mng_total_score']].rename(columns={'mng_total_score':'mng_score'}))
        final_df = fund_indicator_part.join(pd.concat(result)).reset_index().drop_duplicates(subset=['fund_id']).set_index('fund_id')
        final_df = final_df.join(fund_info_part).join(sector_part).join(index_part)
        for fund_id in final_df.index:
            try:
                index_list = sector_fund.loc[[fund_id]].index_id.dropna().unique().tolist()
                if '' in index_list:
                    index_list.remove('')
                append_list = [i for i in index_list if i not in final_df.loc[fund_id,'index_list']]
                final_df.loc[fund_id,'index_list'].extend(append_list)
            except:
                continue
        final_df = final_df.dropna(subset=['total_score'])
        final_df.sector_list = final_df.sector_list.map(lambda x : json.dumps(x))
        final_df.sector_name = final_df.sector_name.map(lambda x : json.dumps(x))
        final_df.index_list = final_df.index_list.map(lambda x : json.dumps(x))
        final_df.index_name = final_df.index_name.map(lambda x : json.dumps(x))
        final_df = final_df.rename(columns={'mdd~history_daily':'mdd','annual_ret~history_m':'annual_ret','name':'mng_name'})
        final_df = final_df.reset_index()
        final_df.loc[:,'is_abs_ret_fund'] = final_df.fund_id.isin(abs_fund_ret_list)
        final_df.loc[:,'is_conv_bond_fund'] = final_df.fund_id.isin(conv_bond_fund_list)
        return final_df
    
    def calc_filter(self, final_df):
        final_df = final_df[final_df.total_score >= 0]
        final_df_1 = final_df[final_df.wind_class_2 != '货币市场型基金']
        final_df_2 = final_df[final_df.wind_class_2 == '货币市场型基金']
        final_df_1 = final_df_1[ (final_df_1.fund_id.isin(self.include_c_type))
                                 & (~ final_df_1.fund_id.isin(self.not_include_in_market_fund_list))
                                 & (~ final_df_1.fund_id.isin(self.ehr_funds))]

        final_df_2 = final_df_2[(~ final_df_2.fund_id.isin(self.not_include_in_market_fund_list))
                                & (~ final_df_2.fund_id.isin(self.not_include_no_buy_funds))]
        final_df = pd.concat([final_df_1, final_df_2])
        fund_size = DataManager.basic_data(func_name='get_fund_size_range')
        fund_size_df = fund_size.pivot_table(columns='fund_id',index='datetime',values='size').tail(1).T
        fund_size_df.columns = ['size']
        big_funds = fund_size_df[fund_size_df['size'] > 1e8].index.tolist()                
        final_df = final_df.loc[final_df.fund_id.isin(big_funds)]
        return final_df

    def calc(self):
        final_df = self.calc_whole_data()
        self.result = self.calc_filter(final_df)

    def insert(self):
        self.append_data(FundRecommendationCollection.__tablename__, self.result)


from ..structure import PositionInfo, OrderStatus
from ..constant import AssetType
from queue import Queue
import datetime
import copy
from pdb import set_trace

class OrderManager:

    def __init__(self,run_engine=None,strategy_type=None):
        self.run_engine = run_engine
        self.positions = {}
        self._cur_id = 0
        self.pending_orders = Queue()
        self.pending_transfers = [] # item is (insert_time, Transfer)
        self.id_orders = {}
        self.trades = {}
        self.ticks = {}
        self.mvs = {}
        self.pos = {}
        self.strategy_type = strategy_type
        self.cash_list = {}

    def init(self):
        pass

    def add_strategy(self, st_name, position_info=None):
        self.positions[st_name] = position_info if position_info else PositionInfo() 
        
        if self.strategy_type == 'stock_daily':
            beginday_yesterday = self.run_engine._data_manager.wrapper.find_beginday_yesterday()
            self.mvs[st_name] = [(beginday_yesterday, position_info.cash.position )] if position_info.cash.position else []
        else:
            self.mvs[st_name] = [] 
        self.trades[st_name] = []
        self.pos[st_name] = {}
        self.cash_list[st_name] = []

    def on_tick(self, tick):
        self.ticks[tick.ticker] = tick
        if self.strategy_type == 'crypto': #TODO 
            for _, position in self.positions.items():
                position.cal_unrealized(tick)

    def check_order(self, order):
        order.order_id = self.new_order_id()
        # 1. record and check availability
        # if has err, set err_msg
        # order.err_msg = 'something wrong'
        order.status = OrderStatus.RESPONDED
        self.id_orders[order.order_id] = order
        self.pending_orders.put(order)
    
    def check_transfer(self, transfer):
        transfer.transfer_id = self.new_order_id()
        # check if this transfer from is ok otherwise refuse
        if not self.get_position(transfer.source).init_transfer(transfer):
            transfer.err_msg = "rejected by PositionInfo"
        else:
            self.pending_transfers.append((self.run_engine.now(), transfer))

    def cancel_order(self, order_id):
        order = self.id_orders.get(order_id)
        if not order is None:
            order.status = OrderStatus.CANCELLED
            self.id_orders.pop(order_id, None)
            return True
        else:
            return self.run_engine._match_engine.cancel_order(order_id)

    def process_orders(self):
        while not self.pending_orders.empty():
            od = self.pending_orders.get()
            self.id_orders.pop(od.order_id, None)
            if od.status == OrderStatus.RESPONDED:
                od.status = OrderStatus.QUEUEING
                self.run_engine._push_order_status(od)
                self.run_engine._match_engine.process_order(od, on_tick=False)

    def process_orders_daily(self, order_list):
        for order in order_list:
            order = copy.deepcopy(order)
            order.order_id = self.new_order_id()
            self.run_engine._match_engine.process_order_daily(order)

    def process_transfers(self, transfer_timedelta=datetime.timedelta(hours=1)):
        cut_time = self.run_engine.now() - transfer_timedelta
        while len(self.pending_transfers) > 0:
            if self.pending_transfers[0][0] > cut_time:
                break
            transfer = self.pending_transfers[0][1]
            self.get_position(transfer.source).confirm_transfer(transfer)
            self.pending_transfers.remove(self.pending_transfers[0])
            self.run_engine._push_transfer(transfer)

    def process_trade(self, trade):
        # 1. record this trade
        # 2. update position
        self.trades[trade.source].append(trade)
        self.positions[trade.source].process_trade(trade)

    def new_order_id(self):
        self._cur_id += 1
        return self._cur_id

    def get_position(self, name):
        return self.positions.get(name)

    def get_trade(self, name):
        return self.trades.get(name)

    def on_day(self, day_data):
        self.day_data = day_data

    def get_market_value(self, name):  
        pos = self.get_position(name) 
        if self.strategy_type  == 'stock_daily':
            mv = pos.cash.position 
            for pi in pos.holdings:
                pt = pi.get_price_ticker()
                if pt in list(self.day_data.keys()):
                    mv += pi.get_market_value(self.day_data[pt])
            return round(mv)
        elif self.strategy_type  == 'crypto': 
            return pos.get_total_market_value() 
        else:
            mv = pos.cash.position 
            for pi in pos.holdings:
                pt = pi.get_price_ticker()
                if self.ticks.get(pt):
                    mv += pi.get_market_value(self.ticks[pt].last_price)
            return mv

    def switch_day(self):
        for st_pos in self.positions:
            for idx, pi in enumerate(self.positions[st_pos].holdings):
                pi.switch_day()
            self.pos[st_pos][self.run_engine.now()] = {pi.ticker : {'volume':pi.position, 'price':self.day_data[pi.ticker]} for pi in self.positions[st_pos].holdings if abs(pi.position) > 1e-9}
        
        cur_dt = self.run_engine.now()
        for st_name in self.run_engine._strategies.keys():
            self.mvs[st_name].append((cur_dt, self.get_market_value(st_name)))
            self.cash_list[st_name].append((cur_dt, self.get_position(st_name).cash.position))
    
    def terminate(self):
        pass

    def print_result(self, st_name):
        print('---------- {} ----------'.format(st_name))
        for dt, mv in self.mvs[st_name]:
            print('{}: {}'.format(dt, mv))
        print('---------- FINISH ----------')
        print('')



import os

from ...util.config import SurfingConfigurator

from .fin_quality_factors import GrossMarginFactor, ReturnOnAssetsFactor, ReturnOnEquityFactor, TurnoverRateOfFAFactor, MixFinQualityFactor
from .growth_factors import IncomeGrowthFactor, NetProfitOfNonRecurringGoLFactor, OperationalCashFlowFactor, MixGrowthFactor
from .leverage_factors import FloatRatioFactor, StockDebtRatioFactor, CashFlowRatioFactor, MixLeverageFactor
from .liquidity_factors import Turnover30DFactor, Turnover60DFactor, Turnover90DFactor, MixLiquidityFactor
from .momentum_factors import AdjMomentum3MFactor, AdjMomentum6MFactor, AdjMomentum12MFactor, MixMomentumFactor
from .scale_factors import MarketValueFloatedFactor
from .tech_factors import ReverseFactor, BiasFactor, RSIFactor, MixTechFactor
from .value_factors import NAToMVFactor, NPToMVFactor, SPSToPFactor, DividendYieldFactor, EBITDAToMVFactor, EBITDAToEVFactor, MixValueFactor
from .volatility_factors import BetaSSEIFactor, Vol250DFactor, ResidualVolSSEIFactor, MixVolatilityFactor
from .fund_basic_factors import FundInfo, InMarketFundInfo, CloseFundInfo, FilteredFundInfoI, TradingDay, IndexInfo, FundManagerInfo, \
                                FundConvInfo, FundBenchmarkInfo, FilteredFundInfoWhole, ActiveFundInfo, IndexCloseDaily, IndexRetDaily, \
                                IndexRetWeekly, IndexRetMonthly, FundNavDaily, FundRetDaily, FundSize, FundSizeCombine, FundCompanyHold, \
                                FundRetDailyModify, FundRetWeeklyModify, FundRetMonthlyModify, FundNavDailyModify, FundInfoAsset, \
                                FundNavWeeklyModify, FundNavMonthlyModify, Macroeco, FundBenchmarkRet, FundBenchmarkPrice, FundPersonalHold
from .fund_derived_factors import TradeYear, FeeRate, AlphaBetaDaily3yI, AlphaDaily3yI, BetaDaily3yI, TrackerrDaily3yI, InfoRatioDaily3yI, \
                                  MddDaily3y, TreynorDaily3yI, DownriskDaily3y, RetOverPeriodDaily3y, AnnualRetDaily3y, AnnualVolDaily3y, \
                                  SharpeDaily3y, AlphaBetaDaily1yI, AlphaDaily1yI, BetaDaily1yI, TrackerrDaily1yI, InfoRatioDaily1yI, \
                                  MddDaily1y, TreynorDaily1yI, DownriskDaily1y, RetOverPeriodDaily1y, AnnualRetDaily1y, AnnualVolDaily1y, \
                                  SharpeDaily1y, AnnualRetDailyHistory, AnnualVolDailyHistory, AnnualRetMonthlyHalfY, AnnualVolMonthlyHalfY, \
                                  SharpeMonthlyHalfY, ContinueRegValue, TotalRetDailyHistory, MddDailyHistory, RecentMonthRet, DownsideStdDailyHistory, \
                                  FundClAlphaBetaHistoryWeekly, FundClAlphaHistoryWeekly, FundClBetaHistoryWeekly, FundClAlphaBeta1YWeekly, FundClAlpha1YWeekly, \
                                  FundClBeta1YWeekly, AlphaMdd3yI, WinRateMonthlyI, WinRateMonthlyTop50, WinRateMonthlyTop75, \
                                  ManagerIndex, ManagerBestFund, ManagerIndexRetDaily, ManagerIndexWeekly, ManagerIndexRetWeekly, \
                                  ManagerIndexMonthly, ManagerIndexRetMonthly, MngAnnualRetDailyHistory, MngAnnualRetDaily1Y, MngAnnualRetDaily3Y, \
                                  MngTotalRetDailyHistory, MngMddDailyHistory, MngAnnualVolDailyHistory, MngDownsideStdDailyHistory, MngDownsideStdMonthlyHistory, \
                                  MngFundTypeTradingDays, MngClAlphaBetaWeekly1Y, MngClAlphaWeekly1Y, MngClBetaWeekly1Y, MngClAlphaBetaWeeklyHistory, \
                                  MngClAlphaWeeklyHistory, MngClBetaWeeklyHistory, MngFundSize

from .fund_derived_score import MngScoreV1, FundMngScoreV1, RetAbilityFundScore, RiskAbilityFundScore, StableAbility, SelectTimeAbility, SelectStockAbility 

conf = SurfingConfigurator().get_aws_settings()
# 在to_parquet/read_parquet中通过storage_options传递如下参数的方法不好用，这里直接设置环境变量
os.environ['AWS_ACCESS_KEY_ID'] = conf.aws_access_key_id
os.environ['AWS_SECRET_ACCESS_KEY'] = conf.aws_secret_access_key
os.environ['AWS_DEFAULT_REGION'] = conf.region_name


for obj in (GrossMarginFactor, ReturnOnAssetsFactor, ReturnOnEquityFactor, TurnoverRateOfFAFactor, MixFinQualityFactor,
            IncomeGrowthFactor, NetProfitOfNonRecurringGoLFactor, OperationalCashFlowFactor, MixGrowthFactor,
            FloatRatioFactor, StockDebtRatioFactor, CashFlowRatioFactor, MixLeverageFactor,
            Turnover30DFactor, Turnover60DFactor, Turnover90DFactor, MixLiquidityFactor,
            AdjMomentum3MFactor, AdjMomentum6MFactor, AdjMomentum12MFactor, MixMomentumFactor,
            MarketValueFloatedFactor,
            ReverseFactor, BiasFactor, RSIFactor, MixTechFactor,
            NAToMVFactor, NPToMVFactor, SPSToPFactor, DividendYieldFactor, EBITDAToMVFactor, EBITDAToEVFactor, MixValueFactor,
            BetaSSEIFactor, Vol250DFactor, ResidualVolSSEIFactor, MixVolatilityFactor):
    obj().register()

fund_basic_factor = [FundInfo, InMarketFundInfo, CloseFundInfo, FilteredFundInfoI, TradingDay, IndexInfo, FundManagerInfo,
                     FundConvInfo, FundBenchmarkInfo, FilteredFundInfoWhole, ActiveFundInfo, IndexCloseDaily, IndexRetDaily,
                     IndexRetWeekly, IndexRetMonthly, FundNavDaily, FundRetDaily, FundSize, FundSizeCombine, FundCompanyHold,
                     FundRetDailyModify, FundInfoAsset, FundRetWeeklyModify, FundRetMonthlyModify, FundNavDailyModify,
                     FundNavWeeklyModify, FundNavMonthlyModify, Macroeco, FundBenchmarkRet, FundBenchmarkPrice, FundPersonalHold]

fund_derived_factor = [TradeYear, FeeRate, AlphaBetaDaily3yI, AlphaDaily3yI, BetaDaily3yI, TrackerrDaily3yI, InfoRatioDaily3yI,
                       MddDaily3y, TreynorDaily3yI, DownriskDaily3y, RetOverPeriodDaily3y, AnnualRetDaily3y, AnnualVolDaily3y,
                       SharpeDaily3y, AlphaBetaDaily1yI, AlphaDaily1yI, BetaDaily1yI, TrackerrDaily1yI, InfoRatioDaily1yI,
                       MddDaily1y, TreynorDaily1yI, DownriskDaily1y, RetOverPeriodDaily1y, AnnualRetDaily1y, AnnualVolDaily1y,
                       SharpeDaily1y, AnnualRetDailyHistory, AnnualVolDailyHistory, AnnualRetMonthlyHalfY, AnnualVolMonthlyHalfY,
                       SharpeMonthlyHalfY, ContinueRegValue, TotalRetDailyHistory, MddDailyHistory, RecentMonthRet, DownsideStdDailyHistory,
                       FundClAlphaBetaHistoryWeekly, FundClAlphaHistoryWeekly, FundClBetaHistoryWeekly, FundClAlphaBeta1YWeekly, FundClAlpha1YWeekly,
                       FundClBeta1YWeekly, ManagerIndex, ManagerBestFund, ManagerIndexRetDaily, ManagerIndexWeekly, ManagerIndexRetWeekly,
                       ManagerIndexMonthly, ManagerIndexRetMonthly, MngAnnualRetDailyHistory, MngAnnualRetDaily1Y, MngAnnualRetDaily3Y,
                       MngTotalRetDailyHistory, MngMddDailyHistory, MngAnnualVolDailyHistory, MngDownsideStdDailyHistory, MngDownsideStdMonthlyHistory,
                       MngFundTypeTradingDays, MngClAlphaBetaWeekly1Y, MngClAlphaWeekly1Y, MngClBetaWeekly1Y, MngClAlphaBetaWeeklyHistory,
                       MngClAlphaWeeklyHistory, MngClBetaWeeklyHistory, MngFundSize, AlphaMdd3yI, WinRateMonthlyI, WinRateMonthlyTop50, WinRateMonthlyTop75]

fund_derived_score = [MngScoreV1,FundMngScoreV1, RetAbilityFundScore, RiskAbilityFundScore, StableAbility, SelectTimeAbility, SelectStockAbility ]

fund_append_update_factor = [AlphaBetaDaily3yI, TrackerrDaily3yI, AlphaBetaDaily1yI, TrackerrDaily1yI, FundClAlphaBetaHistoryWeekly, FundClAlphaBeta1YWeekly,
                             ManagerIndex, MngClAlphaBetaWeekly1Y, MngClAlphaBetaWeeklyHistory, MngScoreV1]



from typing import Dict, List, Set, Optional
import requests
import json
import datetime
import time
import traceback

import pandas as pd

from .config import SurfingConfigurator


class WechatBot(object):

    _SEND_PARTIALLY_UNIT = 4000

    def __init__(self):
        self.wechat_webhook = SurfingConfigurator().get_wechat_webhook_settings('wechat_webhook')

    def send(self, markdown_content):
        print(markdown_content)
        i = 0
        while i < len(markdown_content):
            md_part = markdown_content[i:i+WechatBot._SEND_PARTIALLY_UNIT]
            if i > 0:
                # 非第一部分 加一行注解
                md_part = '**(接上文)**' + md_part
            if i + len(md_part) < len(markdown_content):
                # 如果没办法一次性过把剩余部分发完 倒着找一下换行符
                pos = md_part.rfind('\n')
                if pos != -1:
                    # 如果找到了 这次只发送到这里
                    md_part = md_part[:pos]
                    i += pos
                else:
                    # 否则就直接全发走
                    i += len(md_part)
                # 加一个还有后文的注解
                md_part += '\n**(接下文)**'
            else:
                # 表示这次可以一次性把剩余部分全发完
                i += len(md_part)
            # Send markdown content
            # print(md_part)
            message = {
                'msgtype': 'markdown',
                'markdown': {
                    'content': md_part
                }
            }
            try:
                res = requests.post(url=self.wechat_webhook, data=json.dumps(message), timeout=20)
                print(res)
                time.sleep(4)
            except Exception as e:
                print(e)
                traceback.print_exc()

    def send_data_update_status(self, dt, failed_tasks, updated_count):
        # Send markdown content
        markdown_content = f'{dt} 已完成数据更新，'
        if len(failed_tasks) == 0:
            markdown_content += '所有数据更新成功'
        else:
            markdown_content += '以下数据更新失败: '
            for key, task_names in failed_tasks.items():
                if len(task_names) > 0:
                    markdown_content += f'\n>{key}: <font color=\"error\">{task_names}</font>'

        markdown_content += '\n更新数据统计: '
        for key, item_count in updated_count.items():
            for item, count in item_count.items():
                markdown_content += f'\n>{key}.{item}: <font color=\"info\">{count}</font>'

        return self.send(markdown_content)

    def send_new_fund_list(self, dt, new_funds: Dict[str, str], new_delisted_funds: Dict[str, str], indexes_that_become_invalid: Optional[Set[str]]):
        markdown_content = f'#### {dt}'
        if new_funds is None or new_delisted_funds is None:
            markdown_content += f' **<font color=\"warning\">获取今日新基金或摘牌基金列表失败</font>**'
        else:
            if new_funds:
                markdown_content += f'\n>新基金({len(new_funds)}): <font color=\"info\">{new_funds}</font>'
            else:
                markdown_content += f'\n>无新基金'
            if new_delisted_funds:
                markdown_content += f'\n\n>摘牌基金({len(new_delisted_funds)}): <font color=\"info\">{new_delisted_funds}</font>'
            else:
                markdown_content += f'\n\n>无摘牌基金'
        if indexes_that_become_invalid is not None and indexes_that_become_invalid:
            markdown_content += f'\n\n\n>**失效指数({len(indexes_that_become_invalid)}, 请注意处理!!): <font color=\"warning\">{indexes_that_become_invalid}</font>**'
        self.send(markdown_content)

    def send_fund_list_that_set_end_date_failed(self, dt, fund_list: List[str]):
        markdown_content = f'#### {dt} <font color=\"warning\">设置下列基金摘牌日期失败，请手工修改</font>'
        markdown_content += f'\n>**{fund_list}**'
        self.send(markdown_content)

    def send_untreated_new_fund_list(self, dt, fund_list: List[str]):
        markdown_content = f'#### {dt}'
        if fund_list:
            markdown_content += f'\n>**本周累计有下列新基金未处理其 FundInfo & FundBenchmark({len(fund_list)}): <font color=\"warning\">{fund_list}</font>**'
        else:
            markdown_content += f'\n><font color=\"info\">本周无未处理其 FundInfo & FundBenchmark 的新基金</font>'
        self.send(markdown_content)

    def send_dump_to_s3_result(self, dt, failed_tables: List[str], failed_dfs: List[str]):
        markdown_content = f'#### {dt}'
        if failed_dfs:
            markdown_content += f'\n>**同步下列df失败: <font color=\"warning\">{failed_dfs}</font>**'
        else:
            markdown_content += f'\n><font color=\"info\">同步所有df成功</font>'
        if failed_tables:
            markdown_content += f'\n>**同步下列table失败: <font color=\"warning\">{failed_tables}</font>**'
        else:
            markdown_content += f'\n><font color=\"info\">同步所有table成功</font>'
        self.send(markdown_content)

    def send_update_factor_result(self, dt, tag, failed_factors: List[str]):
        markdown_content = f'#### {dt}'
        if failed_factors:
            markdown_content += f'\n>**下列{tag} factor更新失败: <font color=\"warning\">{failed_factors}</font>**'
        else:
            markdown_content += f'\n><font color=\"info\">所有{tag} factor更新成功</font>'
        self.send(markdown_content)

    def send_hedge_fund_nav_update(self, df: pd.DataFrame):
        markdown_content = '#### <font color=\"info\">以下产品净值数据更新成功</font>'
        markdown_content += f'\n>{df.to_markdown()}'
        self.send(markdown_content)

    def send_hedge_fund_nav_update_failed(self, err_msg: str):
        markdown_content = '#### <font color=\"warning\">产品净值数据更新错误！！</font>'
        markdown_content += f'\n>**err msg:** {err_msg}'
        self.send(markdown_content)

    def send_hedge_fund_info(self, fof_id: str, fund_id: str, fund_name: str, latest_nav_date: datetime.date, fund_data, holding_data, ret_data_with_last_date, ret_date_with_last_pub):
        _FUND_DATA_NAME_DICT = {
            'start_date': '成立日期',
            'days': '成立天数',
            'nav': '最新净值',
            'total_ret': '成立以来收益率',
            'annualized_ret': '年化收益率',
            'annualized_vol': '年化波动率',
            'mdd': '最大回撤',
            'sharpe': '夏普比率',
        }
        _HOLDING_DATA_NAME_DICT = {
            'start_date': '首次买入日期',
            'days': '持有天数',
            'mv': '最新市值',
            'avg_cost': '平均成本',
            'v_nav': '虚拟净值',
            'total_ret': '持仓收益率',
            'annualized_ret': '年化收益率',
            'annualized_vol': '年化波动率',
            'mdd': '最大回撤',
            'sharpe': '夏普比率',
        }

        markdown_content = f'### <font color=\"warning\">{fund_name}</font>({fund_id}) <font color=\"comment\">{latest_nav_date}</font>'
        for k, v in fund_data.items():
            markdown_content += f'\n><font color=\"info\">{_FUND_DATA_NAME_DICT[k]}: </font>**{v}**'
        if ret_data_with_last_date:
            markdown_content += f'\n><font color=\"info\">相对上一个基准日期({ret_data_with_last_date["date"]})收益率: </font>**{ret_data_with_last_date["rr"]}**'
        if ret_date_with_last_pub:
            markdown_content += f'\n><font color=\"info\">相对上一个净值日期({ret_date_with_last_pub["date"]})收益率: </font>**{ret_date_with_last_pub["rr"]}**'
        markdown_content += f'\n>'
        markdown_content += f'\n>**<font color=\"comment\">持仓信息(FOF {fof_id})</font>**'
        for k, v in holding_data.items():
            markdown_content += f'\n><font color=\"info\">{_HOLDING_DATA_NAME_DICT[k]}: </font>**{v}**'
        self.send(markdown_content)

    def send_hedge_fund_info_failed(self, fof_id: str, fund_id: str, err_msg: str):
        markdown_content = f'#### <font color=\"warning\">计算 {fund_id} 信息(FOF: {fof_id})失败</font>'
        markdown_content += f'\n>**err msg:** {err_msg}'
        self.send(markdown_content)

    def send_fof_nav_update(self, df):
        markdown_content = f'#### <font color=\"info\">FOF净值数据更新</font>'
        markdown_content += f'\n>**{df.to_markdown()}**'
        self.send(markdown_content)

    def send_fof_nav_update_failed(self, fof_id: str, err_msg: str):
        markdown_content = f'#### <font color=\"info\">{fof_id}净值数据更新失败！！</font>'
        markdown_content += f'\n>**err msg:** {err_msg}'
        self.send(markdown_content)

    def send_fof_info(self, fof_id: str, latest_nav_date: datetime.date, fund_data):
        _FUND_DATA_NAME_DICT = {
            'start_date': '成立日期',
            'days': '成立天数',
            'mv': '最新市值',
            'nav': '最新净值',
            'total_ret': '成立以来收益率',
            'annualized_ret': '年化收益率',
            'annualized_vol': '年化波动率',
            'mdd': '最大回撤',
            'sharpe': '夏普比率',
        }

        markdown_content = f'### FOF <font color=\"warning\">{fof_id}</font> <font color=\"comment\">{latest_nav_date}</font>'
        # markdown_content += '\n#### <font color=\"comment\">成立以来</font>'
        for k, v in fund_data.items():
            markdown_content += f'\n><font color=\"info\">{_FUND_DATA_NAME_DICT[k]}: </font>**{v}**'
        self.send(markdown_content)

    def send_fof_info_failed(self, fof_id: str, err_msg: str):
        markdown_content = f'#### <font color=\"warning\">计算FOF {fof_id} 信息失败</font>'
        markdown_content += f'\n>**err msg:** {err_msg}'
        self.send(markdown_content)

    def send_research_data_update(self, tag: str, datas: Dict[str, int]):
        markdown_content = f'#### 研究数据({tag})更新成功({len(datas.keys())})'
        for k, v in datas.items():
            markdown_content += f'\n><font color=\"info\">{k}: </font>**{v}**'
        self.send(markdown_content)

    def send_refinancing_key_info(self, dt, stocks, key):
        if stocks:
            markdown_content = f'#### {dt}'
            markdown_content += f'\n下列上市公司发布再融资预案，发行对象中包括 **<font color=\"warning\">{key}</font>**'
            markdown_content += f'\n>**<font color=\"info\">{stocks}</font>**'
            self.send(markdown_content)

    def send_refinancing_impl_key_info(self, dt, stocks, key):
        if stocks:
            markdown_content = f'#### {dt}'
            markdown_content += f'\n下列上市公司发布再融资实施公告，获配对象中包括 **<font color=\"warning\">{key}</font>**'
            markdown_content += f'\n>**<font color=\"info\">{stocks}</font>**'
            self.send(markdown_content)



from cassandra.cqlengine import columns
from cassandra.cqlengine.models import Model
from ...wrapper.cas.cas_connection import CassandraConnector
from ...wrapper.cas.constants import CAS_KEY_SPACE

CassandraConnector().get_conn()

class RealtimeIndexPrice(Model):
    __table_name__ = 'realtime_index_price'
    __keyspace__ = CAS_KEY_SPACE
    __connection__ = 'surfing'

    index_id = columns.Text(primary_key=True)           # 指数ID
    timestamps = columns.List(columns.DateTime())       # 数据时间
    prices = columns.List(columns.Double)               # 指数点位


    def to_dict(self):
        return {
            'index_id': self.index_id,
            'timestamps': self.timestamps,
            'prices': self.prices
        }





import pandas as pd
from datetime import date, datetime, timedelta
from ....api.raw import RawDataApi, RawDatabaseConnector
from .constants import (
    STOCK_INCOME_MAPPING,
    STOCK_BALANCE_MAPPING,
    STOCK_CASH_FLOW_MAPPING,
    STOCK_CASH_FLOW_QUARTERLY_MAPPING,
    STOCK_INCOME_QUARTERLY_MAPPING
)


def read_balance_sheet(stock_id, columns: list):
    reverse_mapping = {v: k for k, v in STOCK_BALANCE_MAPPING.items()}
    columns = [reverse_mapping[i] for i in columns]
    df = RawDataApi().get_em_stock_balance(stock_id, columns)
    df = df.rename(columns=STOCK_BALANCE_MAPPING).set_index('财报日期')
    return df


def read_income_sheet(stock_id, columns: list):
    reverse_mapping = {v: k for k, v in STOCK_INCOME_MAPPING.items()}
    columns = [reverse_mapping[i] for i in columns]
    df = RawDataApi().get_em_stock_income(stock_id, columns)
    df = df.rename(columns=STOCK_INCOME_MAPPING).set_index('财报日期')
    return df


def read_cash_flow_sheet(stock_id, columns: list):
    reverse_mapping = {v: k for k, v in STOCK_CASH_FLOW_MAPPING.items()}
    columns = [reverse_mapping[i] for i in columns]
    df = RawDataApi().get_em_stock_cash_flow(stock_id, columns)
    df = df.rename(columns=STOCK_CASH_FLOW_MAPPING).set_index('财报日期')
    return df


def read_income_quarterly_sheet(stock_id, columns: list):
    reverse_mapping = {v: k for k, v in STOCK_INCOME_QUARTERLY_MAPPING.items()}
    columns = [reverse_mapping[i] for i in columns]
    df = RawDataApi().get_em_stock_income_quarterly(stock_id, columns)
    df = df.rename(columns=STOCK_INCOME_QUARTERLY_MAPPING)
    df = df.drop(columns=['_update_time', 'CODES']).set_index('财报日期')
    return df


def read_cash_flow_quarterly_sheet(stock_id, columns: list):
    reverse_mapping = {v: k for k, v in STOCK_CASH_FLOW_QUARTERLY_MAPPING.items()}
    columns = [reverse_mapping[i] for i in columns]
    df = RawDataApi().get_em_stock_cash_flow_quarterly(stock_id, columns)
    df = df.rename(columns=STOCK_CASH_FLOW_QUARTERLY_MAPPING)
    df = df.drop(columns=['_update_time', 'CODES']).set_index('财报日期')
    return df


def read_stock_dividend(stock_id):
    """
    将分红信息规整到年
    :param stock_id:
    :return: dividend_df
         CODES       DATES DIVPROGRESS  DIVCASHPSBFTAX DIVSTOCKPS DIVCAPITALIZATIONPS DIVPRENOTICEDATE DIVPLANANNCDATE DIVAGMANNCDATE DIVIMPLANNCDATE DIVRECORDDATE   DIVEXDATE  DIVPAYDATE DIVBONUSLISTEDDATE DIVOBJ DIVORNOT        _update_time dividend_type dividend_year
    0  002242.SZ  2018-12-31        实施分配            0.80       None                None             None      2019-03-30     2019-04-23      2019-05-08    2019-05-14  2019-05-15  2019-05-15               None   全体股东        是 2020-08-23 08:00:20             Y    2018-12-31
    1  002242.SZ  2019-06-30        实施分配            0.50       None                None             None      2019-08-15     2019-09-12      2019-09-25    2019-10-09  2019-10-10  2019-10-10               None   全体股东        是 2020-08-23 07:59:41             N    2019-12-31
    2  002242.SZ  2019-12-31        实施分配            0.58       None                None             None      2020-04-01     2020-04-29      2020-05-22    2020-05-28  2020-05-29  2020-05-29               None   全体股东        是 2020-08-23 07:58:43             Y    2019-12-31
    """
    df = RawDataApi().get_em_stock_dividend_by_id(stock_id)
    if len(df) < 1:
        return pd.DataFrame([], columns=['每股股息'])

    df['dividend_type'] = df['DATES'].apply(lambda x: 'Y' if x.strftime('%Y-%m-%d').endswith('12-31') else 'N')
    df['dividend_year'] = df['DATES'].apply(lambda x: date(x.year, 12, 31))

    dff = df[df['dividend_type'] == 'N']
    if len(dff) > 0:
        for index in dff.index:
            temp = df[(df['dividend_year'] == df.loc[index, 'dividend_year']) & (df['dividend_type'] == 'Y')]
            if len(temp) == 0:
                df.loc[index, 'dividend_type'] = 'Y'
            else:
                current_index = temp.index[0]
                df.loc[current_index, 'DIVCASHPSBFTAX'] += df.loc[index, 'DIVCASHPSBFTAX']

    df = df[df['dividend_type'] == 'Y']
    df = df.set_index('DIVEXDATE')
    df = df[['DIVCASHPSBFTAX']]
    df = df.rename(columns={'DIVCASHPSBFTAX': '每股股息'})
    return df


def read_stock_price_sheet(stock_id):
    prices = RawDataApi().get_em_stock_price(
        stock_list=[stock_id],
        columns=['close'],
        start_date=(datetime.now() - timedelta(days=365)).strftime('%Y-%m-%d'),
    ).set_index('datetime')
    prices = prices[['close']]
    prices = prices.rename(columns={'close': '收盘价'})
    return prices


def read_stock_equity_sheet(stock_id):
    empties = RawDataApi().get_em_daily_info(
        stock_list=[stock_id],
        columns=['total_share']
    ).set_index('datetime')
    empties = empties[['total_share']]
    empties = empties.rename(columns={'total_share': '总股本'})
    return empties


def read_financial_real_publish_date(stock_id):
    real_dates = RawDataApi().get_em_stock_fin_fac(
        stock_list=[stock_id],
        columns=['stmt_act_date'],
    ).set_index('datetime')
    real_dates = real_dates[['stmt_act_date']]
    return real_dates




import time
import datetime
import pandas as pd
import numpy as np
from pprint import pprint 
import operator
import traceback
import concurrent.futures
import os
from .backtest import FundBacktestEngine
from .trader import AssetTrader
from ...data.manager.manager_fund import FundDataManager
from ...data.struct import AssetTrade, FundTrade, FundScoreParam, TAAParam, TaaTunerParam, AssetTradeParam, AssetWeight, AssetTimeSpan
from ...data.api.derived import DerivedDataApi
from ...data.manager.score import ScoreFunc

class TaaTuner:

    BACKTEST_NUM = 3
    INDEX_WEIGHT = 0.1
    CASH_WEIGHT = 0.9
    TAA_LOW_TYPE = 'taa_low_only'
    TAA_UP_TYPE = 'taa_up_only'

    def __init__(self, start_time:str='20050101', end_time:str='20200501'):
        self.dm = FundDataManager(start_time=start_time, end_time=end_time)
        self.start_date = datetime.datetime.strptime(start_time, '%Y%m%d').date()
        self.end_date = datetime.datetime.strptime(end_time, '%Y%m%d').date()
        self.ap = AssetTradeParam()
        self.at = AssetTrader(self.ap)

    def init(self, taa_tuner:TaaTunerParam ):
        self.dm.init(score_pre_calc=False)
        self.taa_tuner = taa_tuner
        self.saa = AssetWeight(**{'cash':self.CASH_WEIGHT, self.taa_tuner.index_id:self.INDEX_WEIGHT})
        self.taa_type = self.TAA_LOW_TYPE if self.taa_tuner.IsTaaLowOnly else self.TAA_UP_TYPE
        self.bk_each_time = None

    def grid_search(self):
        self.run_saa()
        self.test_bk_time()
        self.search_bk()
        self.first_round_analysis()
        self.search_bk_2()
        self.second_round_analysis()

    def run_saa(self):
        saa_bk = FundBacktestEngine(data_manager=self.dm, trader=self.at, taa_params=None)
        saa_bk.init()
        saa_bk.run(saa=self.saa, start_date=self.start_date, end_date=self.end_date)
        self.saa_result = saa_bk._report_helper.get_asset_stat()
        
    def search_bk(self, is_test:bool=False, bk_list=None):
        if bk_list is None:
            if is_test:
                taa_param_list = self.taa_tuner.param_list[:self.BACKTEST_NUM]
            else:
                taa_param_list = self.taa_tuner.param_list
        else:
            taa_param_list = bk_list

        self.result = []
        c = 1
        bk_num = len(taa_param_list)
        if self.bk_each_time is not None:
            cost_min = len(taa_param_list) * self.bk_each_time / 60
            print(f'totally bk {bk_num} tasks index_id: {self.taa_tuner.index_id} may cost {cost_min} mins')
        else:
            print(f'totally bk {bk_num} tasks index_id: {self.taa_tuner.index_id}')
        for taa_i in taa_param_list:
            t_0 = time.time()
            taa_bk_i = FundBacktestEngine(data_manager=self.dm, trader=self.at, taa_params=taa_i)
            taa_bk_i.init()
            taa_bk_i.run(saa=self.saa, start_date=self.start_date, end_date=self.end_date)
            taa_result = taa_bk_i._report_helper.get_asset_stat()
            dic = taa_i.__dict__
            dic['taa_ret'] = taa_result['annual_ret']
            self.result.append(dic)
            t_1 = time.time()
            print(f'finish task {c}, cost {t_1 - t_0} sec')
            c += 1
        self.result = pd.DataFrame(self.result)
        self.result['saa_annual_ret'] = self.saa_result['annual_ret']
        self.result['taa_saa_annual_ret'] = self.result['taa_ret'] - self.saa_result['annual_ret']
        self.result['taa_type'] = self.taa_type
        self.result = self.result.sort_values('taa_saa_annual_ret', ascending=False)
        print()

    def test_bk_time(self):
        print()
        print('test backtest')
        t_0 = time.time()
        self.search_bk(is_test=True)
        t_1 = time.time()
        cost_time = t_1 - t_0
        self.bk_each_time = cost_time / self.BACKTEST_NUM
        print(f'bk {self.BACKTEST_NUM} times total cost {cost_time} sec, each bk {self.bk_each_time} sec')
        t_1 = time.time()
        task_num = len(self.taa_tuner.param_list)
        print(f'first round backtest may cost {task_num * self.bk_each_time / 60} mins')
        print()

    def first_round_analysis(self): 
        self.result.to_csv(f'{self.taa_tuner.index_id}_{self.taa_type}_round1.csv')
        self.first_round_result = self.result.iloc[0]
        print('first round filter result')
        pprint(self.first_round_result)
        print()

    def search_bk_2(self):
        d = self.first_round_result 
       
        self.bk_params = []
        if d.taa_type == self.TAA_UP_TYPE:
            LowThreshold = d.LowThreshold
            LowStop = d.LowStop
            LowPlus = d.LowPlus
            for HighMinus in [_ for _ in range( int(d.HighMinus * 100) - 1,int(d.HighMinus * 100) + 2, 1)]:
                for HighStop in [_ for _ in range( int(d.HighStop * 100) - 2,int(d.HighStop * 100) + 3, 1)]:
                    for HighThreshold  in [_ for _ in range( int(d.HighThreshold * 100) - 2,int(d.HighThreshold * 100) + 3, 1)]:
                        t = TAAParam()
                        t.HighThreshold = HighThreshold / 100
                        t.HighStop = HighStop / 100
                        t.HighMinus = HighMinus / 100
                        t.LowStop = LowStop 
                        t.LowThreshold = LowThreshold 
                        t.LowPlus = LowPlus 
                        t.TuneCash = True
                        self.bk_params.append(t)
        else:
            HighThreshold = d.HighThreshold
            HighStop = d.HighStop
            HighMinus = d.HighMinus
            for LowPlus in [_ for _ in range( int(d.LowPlus * 100) - 1,int(d.LowPlus * 100) + 2, 1)]:
                for LowStop in [_ for _ in range( int(d.LowStop * 100) - 2,int(d.LowStop * 100) + 3, 1)]:
                    for LowThreshold in [_ for _ in range( int(d.LowThreshold * 100) - 2,int(d.LowThreshold * 100) + 3, 1)]:
                        t = TAAParam()
                        t.HighThreshold = HighThreshold
                        t.HighStop = HighStop
                        t.HighMinus = HighMinus
                        t.LowStop = LowStop / 100
                        t.LowThreshold = LowThreshold / 100
                        t.LowPlus = LowPlus / 100
                        t.TuneCash = True
                        self.bk_params.append(t)
        self.search_bk(is_test=False, bk_list=self.bk_params)
        
    def second_round_analysis(self):
        self.result.to_csv(f'{self.taa_tuner.index_id}_{self.taa_type}_round2.csv')
        self.second_round_result = self.result.iloc[0]
        print('second round filter result')
        pprint(self.second_round_result)


class SaaTuner:

    ASSET_LIST = ['hs300', 'csi500', 'gem', 'sp500rmb', 'national_debt', 'gold']
    ROUND1_STEP = 5
    ROUND2_STEP = 1
    ROUND1_CSV = 'saa_tuner_round_1.csv'
    ROUND2_CSV = 'saa_tuner_round_2.csv'
    FINAL_CSV = 'saa_tuner_round_final.csv'

    def __init__(self, start_time:str='20050101', end_time:str='20200520'):
        self.dm = FundDataManager(start_time=start_time, end_time=end_time)
        self.start_date = datetime.datetime.strptime(start_time, '%Y%m%d').date()
        self.end_date = datetime.datetime.strptime(end_time, '%Y%m%d').date()
        self.derived = DerivedDataApi()

    def init(self):
        self.result_round1 = []
        self.result_round2 = []
        self.result_target = []
        self.dm.init(score_pre_calc=False)
        self.get_ports_saa()
        self.test_bk_time()
        
    def tune(self):
        self.first_round()
        self.second_round()
        self.final_round()

    def formal_print(self, sentence):
        print(sentence)
        print()

    def get_ports_saa(self):
        asset_info = self.derived.get_asset_allocation_info()
        asset_info = asset_info[['allocation_id','hs300','csi500','gem','sp500rmb','national_debt','gold','cash']].set_index('allocation_id')
        self.port_saa_dic = asset_info.to_dict('index')

    def bk_unit(self, asset:str, amt_diff:float):
        saa = AssetWeight(**{asset:10/100,'cash':90/100})
        asset_param = AssetTradeParam(EnableCommission=True, PurchaseDiscount=0.15, MinActionAmtDiff=amt_diff) 
        t = AssetTrader(asset_param)
        b = FundBacktestEngine(data_manager=self.dm, trader=t, taa_params=None)
        b.init()
        b.run(saa=saa,start_date=self.start_date, end_date=self.end_date)
        res = b.get_asset_result()
        return {'mdd': res['mdd'], 
                'annual_ret': res['annual_ret'], 
                'ret_over_mdd': res['ret_over_mdd'], 
                'asset': asset, 
                'amt_diff': amt_diff, 
                'start_date': self.start_date, 
                'end_date': self.end_date}

    def bk_unit_target(self, amt_diff:float):
        saa = AssetWeight(    
            hs300=15/100,
            csi500=5/100,
            gem=3/100,
            sp500rmb=7/100,
            national_debt=60/100,
            gold=7/100,
            cash=3/100
        )  
        asset_param = AssetTradeParam(EnableCommission=True, PurchaseDiscount=0.15, MinActionAmtDiff=amt_diff) 
        t = AssetTrader(asset_param)
        b = FundBacktestEngine(data_manager=self.dm, trader=t, taa_params=None)
        b.init()
        b.run(saa=saa,start_date=self.start_date, end_date=self.end_date)
        res = b.get_asset_result()
        return {'mdd': res['mdd'], 
                'annual_ret': res['annual_ret'], 
                'ret_over_mdd': res['ret_over_mdd'], 
                'amt_diff': amt_diff, 
                'start_date': self.start_date, 
                'end_date': self.end_date}
    
    def bk_input_saa_amt_diff(self, saa_dict:dict, amt_diff:float, port_id:int=0):
        try:
            saa = saa_dict#AssetWeight(saa_dict)  
            asset_param = AssetTradeParam(EnableCommission=True, PurchaseDiscount=0.15, MinActionAmtDiff=amt_diff) 
            t = AssetTrader(asset_param)
            b = FundBacktestEngine(data_manager=self.dm, trader=t, taa_params=None)
            b.init()
            b.run(saa=saa,start_date=self.start_date, end_date=self.end_date)
            res = b.get_asset_result()
            res['MinActionAmtDiff'] = amt_diff
            res['start_date'] = self.start_date
            res['end_date'] = self.end_date
            res['port_id'] = port_id
            return res
        except Exception as e:
            print(e)
            traceback.print_exc()
            print('boom')
            print(f'port_id {port_id} amt_diff {amt_diff}')
            print(saa_dict)    
            return None
        
    def test_bk_time(self):
        self.formal_print('test backtest')
        t_0 = time.time()
        asset = self.ASSET_LIST[0]
        amt_diff = [i/1000 for i in range(30,40,2)]
        for amt_diff_i in amt_diff:
            self.bk_unit(asset, amt_diff_i)
        t_1 = time.time()
        cost_time = t_1 - t_0
        self.bk_each_time = cost_time / len(amt_diff)
        self.formal_print(f'bk {len(amt_diff)} times total cost {cost_time} sec, each bk {self.bk_each_time} sec')
        
    def first_round(self):
        loop_layer_1 = self.ASSET_LIST
        loop_layer_2 = [_/10000 for _ in range(100, 905, self.ROUND1_STEP)]
        predict_time = len(loop_layer_1) * len(loop_layer_2) * self.bk_each_time / 60
        self.formal_print(f'###first round bk may cost {predict_time} mins')
        bk_time0 = time.time()
        for asset in loop_layer_1:
            asset_t0 = time.time()
            for amt_diff_i in loop_layer_2:
                self.result_round1.append(self.bk_unit(asset, amt_diff_i))
            asset_t1 = time.time()
            cost_m = (asset_t1 - asset_t0) / 60
            self.formal_print(f' - asset {asset} bk finish cost {cost_m} mins')
        bk_time1 = time.time()
        cost_m = (bk_time1 - bk_time0) / 60
        self.formal_print(f'first round finish cost cost {cost_m} mins')
        pd.DataFrame(self.result_round1).to_csv(self.ROUND1_CSV)

    def second_round(self):
        self.round1_df = pd.read_csv(self.ROUND1_CSV, index_col=0)
        loop_layer_1 = self.ASSET_LIST
        layer_2_lens = len([ i for i in range(-self.ROUND1_STEP+1, self.ROUND1_STEP, self.ROUND2_STEP)])
        predict_time = len(loop_layer_1) * layer_2_lens * self.bk_each_time / 60
        self.formal_print(f'###second round bk may cost {predict_time} mins')
        bk_time0 = time.time()
        for asset in loop_layer_1:
            amt_diff_0 = self.round1_df[self.round1_df.asset == asset].sort_values('ret_over_mdd', ascending=False).amt_diff.values[0]
            amt_diff_int = int(amt_diff_0 * 10000)
            for amt_diff_i in range(amt_diff_int - self.ROUND1_STEP + 1, amt_diff_int + self.ROUND1_STEP, self.ROUND2_STEP):
                amt_diff_i = amt_diff_i / 10000
                self.result_round2.append(self.bk_unit(asset, amt_diff_i))
        bk_time1 = time.time()
        cost_m = (bk_time1 - bk_time0) / 60
        self.formal_print(f'second round finish cost cost {cost_m} mins')
        pd.DataFrame(self.result_round2).to_csv(self.ROUND2_CSV)

    def final_round(self):
        self.round2_df = pd.read_csv(self.ROUND2_CSV, index_col=0)
        res = []
        for asset in self.ASSET_LIST:
            res_i = self.round2_df[self.round2_df.asset == asset].sort_values('ret_over_mdd', ascending=False).reset_index(drop=True).iloc[0]
            res.append(res_i)
        self.final_df = pd.DataFrame(res)
        self.formal_print('###final result sort by ret over mdd')
        print(self.final_df)
        self.final_df.to_csv(self.FINAL_CSV)
    
    def target_backtest(self):
        loop_layer = [_/10000 for _ in range(100, 905, self.ROUND1_STEP)]
        for amt_diff_i in loop_layer:
            self.result_target.append(self.bk_unit_target(amt_diff_i))
        self.target_df = pd.DataFrame(self.result_target)
        res = self.target_df.sort_values('ret_over_mdd', ascending=False).reset_index(drop=True).head()
        print(res)

    def bk_fee_search(self, fee_con:bool, MinActionAmtDiff:float):
        saa = AssetWeight(    
            hs300=15/100,
            csi500=5/100,
            gem=3/100,
            sp500rmb=7/100,
            national_debt=60/100,
            gold=7/100,
            cash=3/100
        )  
        asset_param = AssetTradeParam(EnableCommission=fee_con, PurchaseDiscount=0.15, MinActionAmtDiff=MinActionAmtDiff) 
        t = AssetTrader(asset_param)
        b = FundBacktestEngine(data_manager=self.dm, trader=t, taa_params=None)
        b.init()
        b.run(saa=saa,start_date=self.start_date, end_date=self.end_date)
        res = b.get_asset_result()
        res['MinActionAmtDiff'] = MinActionAmtDiff
        res['fee_status'] = fee_con
        res['start_date'] = self.start_date
        res['end_date'] = self.end_date
        return res

    def search_fee(self):
        self.result_fee = []
        #for fee_con in [True, False]:
        fee_con = False
        for MinActionAmtDiff in [0.02,0.03,0.04,0.05,0.06,0.065,0.07,0.08]:
            res_i = self.bk_fee_search(fee_con, MinActionAmtDiff)
            self.result_fee.append(res_i)
        #self.result_fee_df = pd.DataFrame(self.result_fee)

    def bk_unit_double_asset(self, saa, amt_dif):
        asset_param = AssetTradeParam(EnableCommission=True, PurchaseDiscount=0.15, MinActionAmtDiff=amt_dif) 
        t = AssetTrader(asset_param)
        b = FundBacktestEngine(data_manager=self.dm, trader=t, taa_params=None)
        b.init()
        b.run(saa=saa)
        res = b.get_asset_result()
        res['amt_dif'] = amt_dif
        res['rebalance_details'] = b._trader.rebalance_details
        return res

    def bk_loop_selected_amt_diff(self, asset_1, asset_2):
        result = []
        for w in range(10,91,2):
            saa = AssetWeight(**{asset_1:w/100,asset_2:(99-w)/100, 'cash':1/100})
            amt_dif = 50 / 1000
            res_i = self.bk_unit_double_asset(saa=saa, amt_dif=amt_dif)
            result.append(res_i)
        return pd.DataFrame(result)

    def find_trigger_asset(self, dic):
        res = []
        for d, k in dic.items():
            if k:
                res.append({'date':d,'trigger_asset':max(k.items(), key=operator.itemgetter(1))[0]})
        trigger_list = pd.DataFrame(res).trigger_asset.tolist()
        trigger_dic = {i: trigger_list.count(i) for i in set(trigger_list)}
        trigger_asset = max(trigger_dic.items(), key=operator.itemgetter(1))[0]
        return trigger_asset

    def find_trigger_process(self, asset_1, asset_2):
        result = self.bk_loop_selected_amt_diff(asset_1,asset_2)
        res = []
        for idx, r in result.iterrows():
            if len(r.rebalance_date) == 2:
                res.append('unvalid bk, rebalence time <= 2')
                continue
            if idx == 0:
                res.append('None')
                continue
            trigger_asset=self.find_trigger_asset(r.rebalance_details)
            res.append(trigger_asset)
        result['trigger_asset'] = res
        result['rebalance_times'] = [len(_) for _ in result['rebalance_date']]
        return result

class ScoreTuner:

    '''
    收益能力因子:  alpha ret_over_period treynor m_square annual_ret time_ret alpha_w jensen_alpha_m treynor_ratio_m
    均衡能力因子:  info_ratio sharpe calmar_ratio_m information_ratio_m sharpe_ratio_m
    风险能力因子:  beta down_risk mdd var annual_vol beta_w beta_m
    跟踪能力因子:  track_err track_err_w
    费用能力因子:  fee_rate
    '''

    INDICATOR_SIGN={
        'alpha':1,
        'beta':-1,
        'fee_rate':-1,
        'track_err':-1,
        'down_risk':-1,
        'info_ratio':1,
        'ret_over_period':1,
        'treynor':1,
        'mdd':-1,
        'm_square':1,
        'var':-1,
        'r_square':0, #解释 beta alpha 有效性
        'sharpe':1,
        'annual_ret':1,
        'annual_vol':-1,
        'time_ret':1,
        'alpha_w':1,
        'beta_w':-1,
        'track_err_w':-1,
        'beta_m':-1,
        'calmar_ratio_m':1,
        'information_ratio_m':1,
        'jensen_alpha_m':1,
        'sharpe_ratio_m':1,
        'treynor_ratio_m':1,
    }

    def __init__(self, dm:FundDataManager, index_id:str='national_debt', folder:str='good_luck'):
        self.dm = dm
        self.index_id = index_id
        self.folder = folder
        self.search_datelist = list(sorted(set(self.dm.dts.index_fund_indicator_pack.loc[self.index_id].index.get_level_values(0).tolist())))
        self.score_list = self.make_score_list()
        self.index_indicator = self.dm.dts.index_fund_indicator_pack.loc[self.index_id]
        #time_span = AssetTimeSpan().__dict__[index_id]
        #self.index_indicator = self.index_indicator[self.index_indicator.year_length > time_span]
        self.fund_desc_dic = self.dm.dts.fund_info[['fund_id','desc_name']].set_index('fund_id').to_dict()['desc_name']
        self.calculate_std()

    def recent_day(self, dt, duration):
        dts = self.dm.dts.trading_days.datetime
        dts = dts[dts > dt].values
        return dts[0], dts[duration]


    def make_score_equation(self, score_weight:ScoreFunc=ScoreFunc()):
        score_dict = {}
        for indicator, w in score_weight.__dict__.items():
            if w != 0:
                if indicator in ['beta', 'beta_w', 'beta_m']:
                    score_w = self.INDICATOR_SIGN[indicator] * w
                else:
                    score_w = self.INDICATOR_SIGN[indicator] * w / getattr(self.indicator_std_mean, indicator)
                score_i = {indicator: score_w}
                score_dict.update(score_i)
        return ScoreFunc(**score_dict)

    def calculate_std(self):
        dts = self.dm.dts.trading_days.datetime
        dts = dts[dts > datetime.date(2012,1,1)]
        fund_indicator_index = self.dm.dts.fund_indicator[self.dm.dts.fund_indicator.index_id == self.index_id]
        res = []
        for dt in dts:
            std_i = fund_indicator_index[fund_indicator_index.datetime == dt].std(ddof=1)
            res.append(std_i)
        self.indicator_std_mean = pd.DataFrame(res).mean()

    def make_date_param_list(self, time_span:int=242, begin_date=datetime.date(2015,1,1)):
        search_datelist = [_ for _ in self.search_datelist if _ >= begin_date]
        param_list = []
        for idx,d in enumerate(search_datelist):
            end_idx = time_span + idx
            if end_idx < len(search_datelist):
                param_list.append({
                    'begin_date':d,
                    'end_date':search_datelist[end_idx]
                })
        return param_list

    def make_date_param_half_year(self, time_span:int=242, begin_date=datetime.date(2012,1,1), time_lag:int=60):
        search_datelist = [_ for _ in self.search_datelist if _ >= begin_date]
        param_list = []
        for idx,d in enumerate(search_datelist):
            if idx % time_lag != 0:
                continue
            end_idx = time_span + idx
            if end_idx < len(search_datelist):
                param_list.append({
                    'begin_date':d,
                    'end_date':search_datelist[end_idx]
                })
        return param_list

    def make_score_list(self):
        score_list = []
        for alpha in np.arange(0,0.8,0.05):
            for beta in np.arange(0,0.8,0.05):
                for fee_rate in np.arange(0,0.5,0.05):
                    track_err = 1 - alpha - beta - fee_rate
                    if track_err < 0:
                        continue
                    score_list.append(ScoreFunc(alpha=self.INDICATOR_SIGN['alpha'] * alpha, 
                                                beta=self.INDICATOR_SIGN['beta'] * beta, 
                                                fee_rate=self.INDICATOR_SIGN['fee_rate'] * fee_rate,
                                                track_err=self.INDICATOR_SIGN['track_err'] * track_err))
        return score_list

    def fund_alpha_calculator(self,top_num:int=3, 
                                start_date:datetime.date=datetime.date(2018,1,1),
                                end_date:datetime.date=datetime.date(2019,1,1),
                                score_method:ScoreFunc=ScoreFunc(alpha=0.3, beta=-0.6, fee_rate=-0.1)):
        
        index_price = self.dm.dts.index_price.loc[start_date:end_date,[self.index_id]]
        raw_indicator = self.index_indicator.loc[start_date]
        # normalized_indicator = raw_indicator.apply(lambda x: (x - x.mean() + 1e-6)/ (x.std() + 1e-6), axis=0)
        # normalized_indicator['beta'] = raw_indicator['beta']
        normalized_indicator = raw_indicator
        score_raw = normalized_indicator.apply(lambda x: score_method.get(x), axis=1)
        select_funds = pd.DataFrame(score_raw).sort_values(0,ascending=False).index[:top_num].tolist()
        select_nav = self.dm.dts.fund_nav.loc[start_date:end_date,select_funds]
        funds_mean_ret =  np.mean(select_nav.iloc[-1] / select_nav.iloc[0]) - 1
        index_ret = (index_price.iloc[-1] / index_price.iloc[0]).values[0] - 1
        return funds_mean_ret - index_ret

    def select_fund_checker(self,top_num:int=4, 
                                start_date:datetime.date=datetime.date(2018,1,1),
                                end_date:datetime.date=datetime.date(2019,1,1),
                                score_method:ScoreFunc=ScoreFunc(alpha=0.5/ 0.0444, beta=-0.4, track_err=-0.0/0.03939 ,fee_rate=-0.1/ 0.00266)):
        index_price = self.dm.dts.index_price.loc[start_date:end_date,[self.index_id]]
        raw_indicator = self.index_indicator.loc[start_date]
        normalized_indicator = raw_indicator
        score_raw = normalized_indicator.apply(lambda x: score_method.get(x), axis=1)
        select_funds = pd.DataFrame(score_raw).sort_values(0,ascending=False).index[:top_num].tolist()
        select_fund_desc = [self.fund_desc_dic[i] for i in select_funds] 
        
        select_nav = self.dm.dts.fund_nav.loc[start_date:end_date,select_funds]
        funds_mean_ret =  np.mean(select_nav.iloc[-1] / select_nav.iloc[0]) - 1
        index_ret = (index_price.iloc[-1] / index_price.iloc[0]).values[0] - 1
        alpha = funds_mean_ret - index_ret
        return select_fund_desc, alpha
    
    def select_fund_score(self, score_method:ScoreFunc=ScoreFunc(alpha=0.5/ 0.0444, beta=-0.4, track_err=-0.0/0.03939 ,fee_rate=-0.1/ 0.00266)):
        result = []
        param_list = self.make_date_param_half_year()
        for param_i in param_list:
            desc_name, alpha_i = self.select_fund_checker(
                            start_date=param_i['begin_date'], 
                            end_date=param_i['end_date'],
                            score_method=score_method,
                            )
            result.append({
                'datetime' : param_i['begin_date'],
                'alpha' : alpha_i,
                'desc_name': desc_name
            })
            
        alpha_mean = np.mean([i['alpha'] for i in result])
        select_fund = [i['desc_name'] for i in result]
        select_fund = [item for sublist in select_fund for item in sublist]
        power_num = 0
        link_num = 0
        for i in select_fund:
            if '增强' in i:
                power_num += 1
            elif '联接' in i:
                link_num += 1
        power_rate = power_num / len(select_fund)
        link_rate = link_num / len(select_fund)
        dic = {
            '增强比例':power_rate,
            '联接比例':link_rate,
            'alpha':alpha_mean,
            'score_func':score_method,
        }
        return dic
    
    def score_maker(self):
        res = []
        for alpha in range(15,100,3):
            for beta in range(15,100,3):
                fee_rate = 100 - alpha - beta
                if fee_rate < 0:
                    continue
                res.append(ScoreFunc(alpha=alpha/100/0.08291, beta=-beta/100, fee_rate=-fee_rate/100/0.0031)),#csi500
                #hs300 res.append(ScoreFunc(alpha=alpha/100/0.0444, beta=-beta/100, track_err=-0.0/0.03939 ,fee_rate=-fee_rate/100/0.00266))
        return res

    def score_factor_calculator(self, score_method:ScoreFunc=ScoreFunc(alpha=0.3, beta=-0.6, fee_rate=-0.1),
                                    param_idx:int=1):
        t1 = time.time()
        param_list = self.make_date_param_list()
        result = []
        for param_i in param_list:
            alpha_i = self.fund_alpha_calculator(top_num=3, 
                               start_date=param_i['begin_date'], 
                               end_date=param_i['end_date'],
                               score_method=score_method)
            result.append({
                'datetime':param_i['begin_date'],
                'alpha':alpha_i
            })    
        name_list = [f'{fac}_{round(wgt,2)}'for fac, wgt in score_method.__dict__.items() if wgt != 0]
        csv_name = '~'.join(name_list)
        csv_path = self.folder + '/' + str(param_idx) + '-' + csv_name + '.csv'
        pd.DataFrame(result).to_csv(csv_path)
        t2 = time.time()
        print(f'param_idx {param_idx} finish ,cost time {t2 - t1} s')
        return True
        
    def score_checker(self):
        # 日度
        score_list = []
        step = 0.10
        start = 0
        end = 0.5

        for alpha in np.arange(start,end,step):
            for beta in np.arange(start,end,step):
                for ret_over_period in np.arange(start,end,step):
                    for treynor in np.arange(start,end,step):
                        for m_square in np.arange(start,end,step):
                            for annual_ret in np.arange(start,end,step):
                                for time_ret in np.arange(start,end,step):
                                    for jensen_alpha_m in np.arange(start,end,step):
                                        treynor_ratio_m = 1-alpha-beta-ret_over_period-treynor-m_square-annual_ret-time_ret-jensen_alpha_m
                                        if treynor_ratio_m < 0 or treynor_ratio_m > 0.5:
                                            continue
                                        s = ScoreFunc(alpha=alpha,
                                                beta=beta,
                                                ret_over_period=ret_over_period,
                                                treynor=treynor,
                                                m_square=m_square,
                                                annual_ret=annual_ret,
                                                time_ret=time_ret,
                                                jensen_alpha_m=jensen_alpha_m,
                                                treynor_ratio_m=treynor_ratio_m)
                                        score_list.append(self.make_score_equation(s))
        result = []
        for score_i in score_list[:50]:
            res_i = self.select_fund_score(score_i)
            result.append(res_i)
            print(score_list.index(score_i)) 


    def multi_search(self, score_list:list=[ScoreFunc(alpha=0.3, beta=-0.6, fee_rate=-0.1), 
                                            ScoreFunc(alpha=0.3, beta=-0.5, fee_rate=-0.2),
                                            ScoreFunc(alpha=0.2, beta=-0.7, fee_rate=-0.1),
                                            ScoreFunc(alpha=0.1, beta=-0.8, fee_rate=-0.1)],
                           cpu_num:int=2):
        with concurrent.futures.ProcessPoolExecutor(max_workers=cpu_num) as executor:
            future_to_bti = {executor.submit(self.score_factor_calculator, score_method, param_idx): param_idx for param_idx, score_method in enumerate(score_list)}
            for future in concurrent.futures.as_completed(future_to_bti):
                bt_i = future_to_bti[future]
                try:
                    data = future.result()
                except Exception as exc:
                    print(f'{bt_i} generated an exception: {exc}')
                else:
                    pass

    def show_result(self):
        csv_list = os.listdir(self.folder)
        res = []
        for csv_i in csv_list:
            csv_path = self.folder+ csv_i
            df_i = pd.read_csv(csv_path,index_col=0)
            res.append({
                'score_method':csv_i[:-4],
                'alpha_mean':df_i.alpha.mean(),
                'alpha_std':df_i.alpha.std(),
            })
        return pd.DataFrame(res)   





if __name__ == "__main__":
    
    tt = TaaTunerParam()
    tt.IsTaaUpOnly = True
    tt.index_id = 'hs300'

    LowStop = -100
    LowThreshold = -100
    LowPlus = 0

    tt.param_list = []
    for HighThreshold in range(80, 97, 2):
        for HighStop in range(50, 71, 2):
            for HighMinus in [3 , 5, 7]:
                t = TAAParam()
                t.HighThreshold = HighThreshold / 100
                t.HighStop = HighStop / 100
                t.HighMinus = HighMinus / 100
                t.LowStop = LowStop / 100
                t.LowThreshold = LowThreshold / 100
                t.LowPlus = LowPlus / 100
                t.TuneCash = True
                tt.param_list.append(t)

    bk = TaaTuner()
    bk.init(tt)   
    bk.grid_search()

    '''
    asset_1 = 'hs300'
    asset_2 = 'csi500'
    hs300_csi500 = find_trigger_process(asset_1, asset_2)
    hs300_csi500[[asset_1,asset_2,'trigger_asset','rebalance_times']]
    '''



import json
import os
import traceback
from typing import Dict, List, Optional

import pandas as pd
import numpy as np

from .....constant import SourceType
from .....fund.engine.trader import BasicFundTrader
from .....fund.engine.backtest import BasicFundBacktestEngine
from ....fund.basic.basic_data_helper import BasicDataHelper
from ....manager.manager_fund import FundDataManager
from ....view.derived_models import ThirdPartyPortfolioInfo, ThirdPartyPortfolioTrade, ThirdPartyPortfolioIndicator, ThirdPartyPortfolioPositionLatest
from ..derived_data_helper import DerivedDataHelper
from .common import ThirdPortfolioBase


class ManuallyCrawledPortfolioProcessor(ThirdPortfolioBase):

    _BLOCK_LIST = [
    ]

    _SRC_TO_NAME = {
        'bcm': '交通银行',
        'ccb': '建设银行',
        'cgb': '广发银行',
        'citics': '中信证券',
        'citicb': '中信银行',
        'cmbc': '民生银行',
        'gtjas': '国泰君安',
        'guosens': '国信证券',
        'icbc': '工商银行',
        'pinganb': '平安银行',
        'spdb': '浦发银行',
    }

    def __init__(self, data_helper: DerivedDataHelper):
        super().__init__()

        self._data_helper = DerivedDataHelper()
        self._bdh = BasicDataHelper()
        self._params_json: Dict[str, str] = {}
        self._dir_path: str = f'{os.path.dirname(os.path.abspath(__file__))}/data/manually_crawled'
        self._trades_df_list: List[pd.DataFrame] = []
        self._nav: Dict[str, pd.DataFrame] = {}
        self._last_position: List[pd.Series] = []

    def init(self):
        # self._fdm = FundDataManager(s3_retriever_activation=True)
        # self._fdm.init(score_pre_calc=True, print_time=True)

        from .....resource.data_store import DataStore
        self._fdm: FundDataManager = DataStore.load_dm()
        self._trading_day: pd.DataFrame = FundDataManager.basic_data('get_trading_day_list').drop(columns='_update_time')

    def _calc_one(self, po_src: str, po_path: str):
        from ....view.cas.third_party_portfolio_nav import ThirdPartyPortfolioNav

        def transform_fund_id(x: pd.Series) -> Optional[str]:
            fund_id = self._bdh._get_fund_id_from_order_book_id(x.fund_id, x.trading_day)
            if fund_id is None:
                print(f'[WARNING] can not fund id {x.fund_id} (po_id){po_id} (po_name){po_name} (trading_day){x.trading_day}')
            return fund_id

        def calc_fund_weight_list(x) -> pd.Series:
            weights_sum = 0
            fund_weights = []
            for row in x.itertuples(index=False):
                fund_weights.append([row.fund_id, row.fund_weight])
                weights_sum += row.fund_weight
            weights_sum = round(weights_sum, 8)
            if weights_sum != 1:
                print(f'sum of percent not equals 1 (po_id){po_id} (po_name){po_name} (trade_date){x.trading_day.array[0]} (weights_sum){weights_sum}')
            return pd.Series({'fund_weight_list': fund_weights})

        def calc_fund_trade_list(x) -> pd.Series:
            fund_trades = []
            for row in x.itertuples(index=False):
                fund_trades.append({'fund_id': row.fund_id, 'from_percent': row.before_w, 'to_percent': row.after_w})
            return pd.Series({'po_id': po_id, 'details': json.dumps(fund_trades)})

        # 几个list，把当前目录下每个组合的几部分数据分别存储起来
        basic_info: List[pd.Series] = []
        trades_df_list: List[pd.DataFrame] = []
        last_positions: List[pd.Series] = []
        indicators: List[pd.Series] = []
        number = 1
        with os.scandir(po_path) as it:
            for entry in it:
                if not entry.is_file() or not entry.name.endswith('.csv'):
                    continue
                # 取文件名并且只取'_'前边的
                po_name = entry.name.split('.')[0].split('_')[0]
                # 转成大写后加一个6位数字，来生成一个组合ID
                po_id = po_src.upper() + str(number).zfill(6)
                number += 1
                # 这里如果找不到对应的枚举类型则需要去添加一下
                real_po_src = SourceType[po_src]
                # 根据英语简称找到中文的主理人名称
                manager_name = self._SRC_TO_NAME[po_src]

                # 读取调仓文件
                df = pd.read_csv(entry.path, dtype={'trading_day': 'str', 'fund_id': 'str'})
                # 调仓文件中最小的日期即为成立日期
                start_date = pd.to_datetime(df.trading_day.min(), infer_datetime_format=True).date()
                # 组合起来基础信息
                basic_info.append(pd.Series({'po_name': po_name, 'po_id': po_id, 'po_src': real_po_src, 'manager_name': manager_name, 'start_date': start_date.isoformat()}))

                df['trading_day'] = pd.to_datetime(df.trading_day, infer_datetime_format=True).dt.date
                # 转换成我们自己的fund_id
                df['fund_id'] = df.apply(transform_fund_id, axis=1)
                # 以trading day为单位来生成fund weight list
                fund_weight_list = df.groupby(by='trading_day', sort=False).apply(calc_fund_weight_list)
                # 我们是第二天调仓，因此这里调整为之前的第一个交易日
                fund_weight_list = [[self._trading_day[self._trading_day.datetime < row.Index].datetime.array[-1], row.fund_weight_list] for row in fund_weight_list.itertuples()]
                # 按日期升序排序
                fund_weight_list.sort(key=lambda x: x[0])
                # 我们是第二天调仓，因此这里调整为之前的第一个交易日
                start_date = self._trading_day[self._trading_day.datetime < start_date].datetime.array[-1]
                t = BasicFundTrader()
                b = BasicFundBacktestEngine(data_manager=self._fdm, trader=t, fund_weight_list=fund_weight_list)
                b.init()
                b.run(start_date=start_date, end_date=None, print_time=False)

                fund_trade = b.get_fund_trade()
                if not fund_trade.empty:
                    # 以trade day为单位来获取实际调仓记录
                    trades_df_list.append(fund_trade.groupby(by='trade_date', sort=False).apply(calc_fund_trade_list).reset_index().rename(columns={'trade_date': 'datetime'}))
                else:
                    print(f'[WARNING] got a empty fund trade (po_id){po_id} (po_name){po_name} (po_src){po_src}')

                # 获取最新持仓
                last_position: pd.DataFrame = b.get_last_position()
                if last_position is not None:
                    last_positions.append(pd.Series({'po_id': po_id, 'datetime': last_position.index.name, 'po_src': real_po_src, 'position': last_position.loc[:, 'weight'].to_json()}))
                else:
                    print(f'[WARNING] got a empty last positions (po_id){po_id} (po_name){po_name} (po_src){po_src}')

                result = b.get_fund_result()
                if result is None:
                    continue
                # 只保留我们需要的回测结果
                result = {k: v for k, v in result.items() if k in ('market_value', 'last_unit_nav', 'mdd', 'annual_ret', 'mdd_d1', 'mdd_d2', 'total_commission', 'rebalance_times')}
                result['datetime'] = result['market_value'].index.array[-1]
                result['from_setup_ret'] = result['last_unit_nav'] - 1
                resampled = result['market_value'].copy()
                resampled.index = pd.to_datetime(resampled.index)
                resampled_w = resampled.mv.resample('1W').apply(lambda x: x[-1] if not x.empty else np.nan)
                result['vol_by_week'] = resampled_w.dropna().pct_change().std(ddof=1) * np.sqrt(52)

                resampled_m = resampled.mv.resample('1M').apply(lambda x: x[-1] if not x.empty else np.nan)
                result['vol_by_month'] = resampled_m.dropna().pct_change().std(ddof=1) * np.sqrt(12)
                result['vol_by_day'] = resampled.mv.dropna().pct_change().std(ddof=1) * np.sqrt(244)
                result['annual_compounded_ret'] = np.power(result['last_unit_nav'], 1 / (result['market_value'].shape[0] / 244)) - 1
                if result['vol_by_week'] != 0:
                    result['sharpe_ratio'] = (result['annual_compounded_ret'] - 0.02) / result['vol_by_week']
                else:
                    result['sharpe_ratio'] = np.nan

                # 遍历每个需要计算的时间区间
                for name, value in self._DIM_TYPE.items():
                    bd = (result['datetime'] - value).date()
                    # 试图寻找bd及其之前的最近一天
                    filtered_mv = result['market_value'].loc[result['market_value'].index <= bd, :]
                    if filtered_mv.empty:
                        continue
                    # 计算区间内的return
                    result[name] = result['market_value'].mv.array[-1] / filtered_mv.mv.array[-1] - 1

                result['nav'] = result['last_unit_nav']
                result['po_id'] = po_id
                self._nav[po_id] = result['market_value'].mv / result['market_value'].mv.array[0]
                self._nav[po_id].index = self._nav[po_id].index.map(lambda x: x.isoformat())
                # 删掉我们不需要的回测结果
                for one in ('market_value', 'last_unit_nav'):
                    del result[one]
                indicators.append(pd.Series(result))

        print(pd.DataFrame(basic_info))
        self._data_helper._upload_derived(pd.DataFrame(basic_info), ThirdPartyPortfolioInfo.__table__.name)
        if trades_df_list:
            trades = pd.concat(trades_df_list)
            trades['po_src'] = real_po_src
            print(trades)
            self._data_helper._upload_derived(trades, ThirdPartyPortfolioTrade.__table__.name)
        else:
            print(f'empty trades of {po_src}')

        indicators = pd.DataFrame(indicators)
        indicators['po_src'] = real_po_src
        print(indicators)
        self._data_helper._upload_derived(indicators, ThirdPartyPortfolioIndicator.__table__.name)

        print(pd.DataFrame(last_positions))
        self._data_helper._upload_derived(pd.DataFrame(last_positions), ThirdPartyPortfolioPositionLatest.__table__.name)

        # nav存cas
        # from cassandra.cqlengine.query import BatchQuery
        # with BatchQuery() as batch:
        #     for po_id, mv in self._nav.items():
        #         ThirdPartyPortfolioNav.objects(po_id=po_id).batch(batch).update(
        #             date__append=mv.index.to_list(),
        #             nav__append=mv.to_list(),
        #         )
        for po_id, mv in self._nav.items():
            ThirdPartyPortfolioNav.create(
                po_id=po_id,
                date=mv.index.to_list(),
                nav=mv.to_list(),
            )

    def calc(self):
        # 遍历每个手工爬取的目录
        with os.scandir(self._dir_path) as it:
            for entry in it:
                if entry.is_dir():
                    # 计算每个目录下的组合
                    print(f'calc {entry.name}')
                    self._calc_one(entry.name, entry.path)
                    print(f'{entry.name} done')

    def process(self, end_date: str) -> List[str]:
        failed_tasks: List[str] = []
        try:
            po_src_list = []
            for one in SourceType:
                if one >= SourceType.bcm:
                    po_src_list.append(one.value)
            self.calc_everyday(end_date, po_src_list)
        except Exception as e:
            print(e)
            traceback.print_exc()
            failed_tasks.append('third_party_portfolio_manually_crawled')
        return failed_tasks


if __name__ == '__main__':
    mc_pp = ManuallyCrawledPortfolioProcessor(DerivedDataHelper())
    # mc_pp.init()
    # mc_pp.calc()
    # mc_pp.process('20201012')




import datetime
from typing import Optional, List, Union

import pandas as pd
import numpy as np
from pandas.tseries.offsets import DateOffset
from sklearn.preprocessing import StandardScaler

from ...constant import StockFactorType
from ...data.manager import DataManager
from .base import Factor


_UTILS_PREFIX = 'utils/'
_START_DATE = '2005-01-01'

_TRADING_DAYS_PER_YEAR = 242


def get_day(num, date_list, last_date):
    value = DateOffset(days=num)
    # 根据时间区间计算出begin date(bd)
    bd = (last_date - value).date()
    # 试图寻找bd及其之前的最近一天
    bd_list = date_list[date_list <= bd]
    return bd_list[-1]


def get_quarterly_tradingday_list() -> List[datetime.date]:
    tradingday = TradingDayList().get()
    tradingday_list: List[datetime.date] = []
    for date in zip(tradingday, tradingday.shift(1)):
        if np.nan in date:
            continue
        today = date[1]
        next_trading_dt = date[0]
        if today.year == next_trading_dt.year and today.month == next_trading_dt.month:
            continue
        if today.month not in (3, 6, 9, 12):
            continue
        tradingday_list.append(today)
    return tradingday_list


def get_factor_from_stock_data(table: str, columns: List[str], index='datetime', processor_after_pivot=None):
    df: Optional[pd.DataFrame] = DataManager.raw_data(table, columns=columns)
    df = df.pivot(index=index, columns='stock_id', values=columns)
    if len(columns) == 1:
        df = df[columns[0]]
    if processor_after_pivot is not None:
        df = processor_after_pivot(df)

    trading_day_list = TradingDayList().get()
    return df.reindex(df.index.union(trading_day_list)).ffill().reindex(trading_day_list)


def calc_growth_rate(x, whole_df):
    date_begin = x.name - DateOffset(years=1)
    date_list = whole_df.index[whole_df.index <= date_begin]
    if date_list.empty:
        return
    return whole_df.loc[x.name, :] / whole_df.loc[date_list.array[-1], :] - 1


def normalize(a: Union[pd.DataFrame, pd.Series]) -> np.ndarray:
    if isinstance(a, pd.Series):
        a = pd.DataFrame(a)
    return StandardScaler().fit_transform(a.replace({np.Inf: np.nan, -np.Inf: np.nan}))


def calc_ttm(x, whole_df):
    date = x.name
    if date.month == 12 and date.day == 31:
        return x
    else:
        try:
            last_annual_report = whole_df.loc[datetime.date(date.year-1, 12, 31), :]
            same_period_last_year = whole_df.loc[datetime.date(date.year-1, date.month, date.day), :]
            return x + last_annual_report - same_period_last_year
        except KeyError:
            return np.nan


class TradingDayList(Factor):

    def __init__(self):
        super().__init__(_UTILS_PREFIX + 'td_list', StockFactorType.BASIC)

    def calc(self):
        trading_day_list: Optional[pd.DataFrame] = DataManager.basic_data('get_trading_day_list', start_date=_START_DATE)

        self._factor = trading_day_list.datetime

    def trim_by_date(self, end_date: str):
        if self._factor is None:
            return
        self._factor = self._factor[self._factor <= pd.to_datetime(end_date).date()]


class FundHoldStockFactor(Factor):

    def __init__(self):
        super().__init__(_UTILS_PREFIX + 'fund_hold_stock', StockFactorType.BASIC)

    def calc(self):
        self._factor = DataManager.basic_data('get_fund_hold_stock_by_id').drop(columns='_update_time')


class StockIndustryInfoFactor(Factor):

    def __init__(self):
        super().__init__(_UTILS_PREFIX + 'stock_ind_info', StockFactorType.BASIC)

    def calc(self):
        df: Optional[pd.DataFrame] = DataManager.raw_data('get_em_stock_info').drop(columns='_update_time')
        self._factor = df.set_index('stock_id').bl_sws_ind_code.str.split(pat='-', expand=True).iloc[:, -1]



import dataclasses
import json

@dataclasses.dataclass
class WxSeqItem(object):
    seq: int = None
    msgid: str = None
    publickey_ver: int = None
    encrypt_random_key: str = None
    encrypt_chat_msg: str = None

    def __repr__(self):
        return f'<Seq seq={self.seq} msg={self.msgid} ver={self.publickey_ver}>'

@dataclasses.dataclass
class WxRawData(object):
    # input
    errcode: int = None
    errmsg: str = None
    chatdata: list = None

    def __post_init__(self):
        if self.chatdata:
            for i, msg in enumerate(self.chatdata):
                if isinstance(msg, dict):
                    self.chatdata[i] = WxSeqItem(**msg)

    @property
    def max_seq(self):
        return self.chatdata[-1].seq if len(self.chatdata) > 0 else None

    @property
    def min_seq(self):
        return self.chatdata[0].seq if len(self.chatdata) > 0 else None

    def __repr__(self):
        return f'<Data errcode={self.errcode} errmsg={self.errmsg} chatdata(len={len(self.chatdata)} min_seq={self.min_seq} max_seq={self.max_seq})>'

@dataclasses.dataclass
class WxMsgItem(object):
    msgid: str = None
    action: str = None # send
    from_id: str = None
    to_ids: list = None
    roomid: str = None # only for room
    msgtime: int = None # timestamp: 1585113286223
    msgtype: str = None # text / ...
    text: dict = None # text_raw
    picture: dict = None # picture
    info: dict = None # markdown
    text_content: str = None # text_content
    md_content: str = None # text_content

    # analyze
    scenario_id: int = None # 识别出来的场景ID
    keywords: list = None   # 场景相关的关键词
    mood_level: int = None  # 情绪指标

    def __init__(self, decrypted: dict):
        self.msgid = decrypted.get('msgid')
        self.action = decrypted.get('action')
        self.from_id = decrypted.get('from')
        self.to_ids = decrypted.get('tolist')
        self.roomid = decrypted.get('roomid')
        self.msgtime = decrypted.get('msgtime')
        self.msgtype = decrypted.get('msgtype')
        self.text = decrypted.get('text')
        self.info = decrypted.get('info')
        if self.text:
            self.text_content = self.text.get('content')
        if self.info:
            self.md_content = self.info.get('content')
    
    def __repr__(self):
        return f'<Msg type={self.msgtype} from={self.from_id} to={self.to_ids} msg={self.text_content}>'


import pandas as pd
import glob
from ...wrapper.mysql import RawDatabaseConnector

def hf_fund_info(f):
    #f = '/Users/huangkejia/Downloads/pics/'
    csv = '跟踪产品.xlsx'
    sheet = '产品列表'
    df = pd.read_excel(f+csv,sheet_name=sheet)
    df.columns = df.iloc[0, :]
    df.drop(df.index[0], inplace=True)
    cols = df.columns.tolist()
    col_dic = {c: c.replace('*','') for c in cols}
    df = df.rename(columns=col_dic)
    td = pd.to_datetime(df.成立日期)
    td = [i.date() for i in td ]
    df= df[['策略类型','产品简称','产品ID','成立日期','机构ID','机构简称','基金经理','来源','净值来源']]
    dic = {
        '策略类型':'stg_type',
        '产品简称':'desc_name',
        '产品ID':'fund_id',
        '成立日期':'begin_date',
        '机构ID':'agency_id',
        '机构简称':'agency_name',
        '基金经理':'manager_name',
        '来源':'source',
        '净值来源':'nav_source',
    }
    df = df.rename(columns=dic)
    df.to_sql('hf_fund_info', RawDatabaseConnector().get_engine(), index=False, if_exists='append')

def hf_fund_nav(f):
    #f = '/Users/huangkejia/Downloads/pics/'
    csv = '跟踪产品.xlsx'
    sheet = '基金净值'
    df = pd.read_excel(f+csv,sheet_name=sheet)
    df = df.set_index('datetime')
    df = pd.DataFrame(df.stack()).reset_index()
    df = df.rename(columns={'level_1':'fund_id',0:'nav'})
    df.to_sql('hf_fund_nav', RawDatabaseConnector().get_engine(), index=False, if_exists='append')

def hf_index_price(f):
    #f = '/Users/huangkejia/Downloads/pics/'
    file_list = glob.glob(f+'*.xlsx')
    ls = ['/Users/huangkejia/Downloads/pics/跟踪产品.xlsx','/Users/huangkejia/Downloads/pics/全产品净值列表.xlsx']
    csv_list = [i for i in file_list if i not in ls]
    dics = {   
        'stg_longonly':'融智-股票策略-主观多头指数',
        'stg_indexpow':'融智-股票策略-量化多头指数',
        'stg_neomark':'融智-相对价值-股票市场中性指',
        'stg_future':'融智-管理期货-管理期货复合指数',
        'stg_hedge':'融智-股票策略-股票多空指数',
        'stg_timdriven':'朝阳永续-事件驱动精选指数',
        'stg_macrohedge':'融智-宏观策略指数',
        'stg_highfreq':'融智-相对价值-套利指数',
        'stg_bond':'融智-固定收益-固收复合指数',
        'stg_fof':'融智-组合基金指数',
    }
    dics_revs = {v:k for k,v in dics.items()}
    index_result = []
    for csv_i in csv_list:
        df = pd.read_excel(csv_i)
        df.columns = df.iloc[2, :]
        df.drop(df.index[0:3], inplace=True)
        _cols = df.columns.tolist()
        df = df.drop(columns=[_cols[3]])
        name_dic = {'指数日期':'index_date','指数计算日期':'calcu_date',_cols[2]:'close'}
        df = df.rename(columns=name_dic)
        for name, code_i in dics_revs.items():
            if name in csv_i:
                code = code_i
                break
        df.loc[:,'index_id'] = code    
        index_result.append(df)
    df_i = '/Users/huangkejia/Downloads/pics/朝阳永续-事件驱动精选指数.xls'
    df = pd.read_excel(df_i)

    _dic = {
        '时间':'index_date',
        '事件驱动精选指数':'close',
    }
    df = df[list(_dic.keys())].replace('--',None).dropna().rename(columns=_dic)
    df.loc[:,'calcu_date'] = df.index_date
    df.loc[:,'index_id'] = 'stg_timdriven'
    index_result.append(df)
    df_result = pd.concat(index_result)
    num = df_result.close.tolist()
    num = [i if isinstance(i, float) else float(i.replace(',','')) for i in num]
    df_result.close = num
    df_result.to_sql('hf_index_price', RawDatabaseConnector().get_engine(), index=False, if_exists='append')


import traceback
from typing import List

import pandas as pd
import numpy as np
import statsmodels.api as sm
from sklearn.preprocessing import StandardScaler

from ...api.basic import BasicDataApi
from ...view.derived_models import StyleReg
from .derived_data_helper import DerivedDataHelper

class StyleBoxReg:
    
    def __init__(self,_data_helper: DerivedDataHelper):
        self._data_helper: DerivedDataHelper = _data_helper
        self._basic_api = BasicDataApi()
        
    def init(self, start_date, end_date):
        self.start_date = start_date
        self.end_date = end_date
        #从db读取六个指数的数据
        self._large_v_price = self._basic_api.get_index_price(['cni_largec_v'])
        self._large_g_price = self._basic_api.get_index_price(['cni_largec_g'])
        self._mid_v_price = self._basic_api.get_index_price(['cni_midc_v'])
        self._mid_g_price = self._basic_api.get_index_price(['cni_midc_g'])
        self._small_v_price = self._basic_api.get_index_price(['cni_smallc_v'])
        self._small_g_price = self._basic_api.get_index_price(['cni_smallc_g'])
        #db表格转收益率序列
        def get_return(test, start_date, end_date):
            start_date_dt = pd.to_datetime(start_date, infer_datetime_format=True)
            end_date_dt = pd.to_datetime(end_date, infer_datetime_format=True)
            test = test[['datetime','close']]
            test = test.set_index(test.datetime)
            temp = test.close
            temp = temp.apply(np.log).diff()[1:]
            temp = temp[np.logical_and(temp.index>=start_date_dt, temp.index<=end_date_dt)]
            return temp
        self.large_v_return = get_return(self._large_v_price, start_date, end_date)
        self.large_g_return = get_return(self._large_g_price, start_date, end_date)
        self.mid_v_return = get_return(self._mid_v_price, start_date, end_date)
        self.mid_g_return = get_return(self._mid_g_price, start_date, end_date)
        self.small_v_return = get_return(self._small_v_price, start_date, end_date)
        self.small_g_return = get_return(self._small_g_price, start_date, end_date)
        self.x = pd.concat([self.large_g_return, self.large_v_return, self.mid_g_return, self.mid_v_return, self.small_g_return, self.small_v_return], axis = 1)
        self.x.columns = ['large_g','large_v','mid_g','mid_v','small_g','small_v']
    
    def process(self, start_date, end_date):
        failed_tasks: List[str] = []
        try:
            self.init(start_date=start_date, end_date=end_date)
            self._data_helper._upload_derived(self.x.reset_index(), StyleReg.__table__.name)
        except Exception as e:
            print(e)
            traceback.print_exc()
            failed_tasks.append(StyleReg.__table__.name)
        return failed_tasks
        
    
    @staticmethod
    def cal_fund(factor_return: pd.DataFrame, fund_nav:pd.Series):
        fund_ret = np.log(fund_nav).diff().iloc[1:].rename('fund_ret')
        whole_data = factor_return.join(fund_ret)
        whole_data = whole_data[whole_data.notna().all(axis=1)]
        X = sm.add_constant(whole_data.drop(columns='fund_ret'))
        Y = whole_data.fund_ret
        model = sm.OLS(Y, X)
        resu = model.fit()
        rets = (resu.params[1:] * whole_data.drop(columns='fund_ret'))
        summation = rets.sum(axis=1)
        return (resu.params, whole_data.loc[:, 'fund_ret'] - summation, whole_data.fund_ret, rets)
    
    @staticmethod
    def fund_analysis_plot(fund_ret: pd.Series, factor_ret_contri: pd.DataFrame):
        import matplotlib.pyplot as plt

        total_data = pd.concat([factor_ret_contri, fund_ret], axis=1)
        summation = total_data.drop(columns='fund_ret').sum(axis=1)
        total_data['spec'] = total_data.fund_ret - summation
        total_data.cumsum().plot(figsize=(18, 12), grid=True, title='基金收益率分解', colormap=plt.cm.get_cmap('nipy_spectral'))
        plt.legend(fontsize=15)


if __name__ == '__main__':
    sbg = StyleBoxReg(DerivedDataHelper())
    sbg.process('20200922', '20200922')
